from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

from sae_bench.evals.base_eval_output import (
    DEFAULT_DISPLAY,
    BaseEvalOutput,
    BaseMetricCategories,
    BaseMetrics,
    BaseResultDetail,
)
from sae_bench.evals.scr_and_tpp.eval_config import ScrAndTppEvalConfig

EVAL_TYPE_ID_SCR = "scr"
EVAL_TYPE_ID_TPP = "tpp"


@dataclass
class ScrMetrics(BaseMetrics):
    scr_dir1_threshold_2: float | None = Field(
        None,
        title="SCR Dir 1, Top 2 SAE latents",
        description="Ablating the top 2 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_2: float | None = Field(
        None,
        title="SCR Metric, Top 2 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 2 SAE latents",
    )
    scr_dir2_threshold_2: float | None = Field(
        None,
        title="SCR Dir 2, Top 2 SAE latents",
        description="Ablating the top 2 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_5: float | None = Field(
        None,
        title="SCR Dir 1, Top 5 SAE latents",
        description="Ablating the top 5 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_5: float | None = Field(
        None,
        title="SCR Metric, Top 5 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 5 SAE latents",
    )
    scr_dir2_threshold_5: float | None = Field(
        None,
        title="SCR Dir 2, Top 5 SAE latents",
        description="Ablating the top 5 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_10: float | None = Field(
        None,
        title="SCR Dir 1, Top 10 SAE latents",
        description="Ablating the top 10 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_10: float | None = Field(
        None,
        title="SCR Metric, Top 10 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 10 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    scr_dir2_threshold_10: float | None = Field(
        None,
        title="SCR Dir 2, Top 10 SAE latents",
        description="Ablating the top 10 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_20: float | None = Field(
        None,
        title="SCR Dir 1, Top 20 SAE latents",
        description="Ablating the top 20 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_20: float | None = Field(
        None,
        title="SCR Metric, Top 20 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 20 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    scr_dir2_threshold_20: float | None = Field(
        None,
        title="SCR Dir 2, Top 20 SAE latents",
        description="Ablating the top 20 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_50: float | None = Field(
        None,
        title="SCR Dir 1, Top 50 SAE latents",
        description="Ablating the top 50 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_50: float | None = Field(
        None,
        title="SCR Metric, Top 50 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 50 SAE latents",
    )
    scr_dir2_threshold_50: float | None = Field(
        None,
        title="SCR Dir 2, Top 50 SAE latents",
        description="Ablating the top 50 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_100: float | None = Field(
        None,
        title="SCR Dir 1, Top 100 SAE latents",
        description="Ablating the top 100 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_100: float | None = Field(
        None,
        title="SCR Metric, Top 100 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 100 SAE latents",
    )
    scr_dir2_threshold_100: float | None = Field(
        None,
        title="SCR Dir 2, Top 100 SAE latents",
        description="Ablating the top 100 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_500: float | None = Field(
        None,
        title="SCR Dir 1, Top 500 SAE latents",
        description="Ablating the top 500 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_500: float | None = Field(
        None,
        title="SCR Metric, Top 500 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 500 SAE latents",
    )
    scr_dir2_threshold_500: float | None = Field(
        None,
        title="SCR Dir 2, Top 500 SAE latents",
        description="Ablating the top 500 profession latents to increase gender accuracy",
    )


@dataclass
class ScrMetricCategories(BaseMetricCategories):
    scr_metrics: ScrMetrics = Field(
        title="SCR Metrics",
        description="SCR metrics, calculated for different numbers of ablated features. Also includes the results for both correlation removal directions.",
        json_schema_extra=DEFAULT_DISPLAY,
    )


@dataclass
class ScrResultDetail(BaseResultDetail):
    dataset_name: str = Field(title="Dataset Name", description="")

    scr_dir1_threshold_2: float | None = Field(
        None,
        title="SCR Dir 1, Top 2 SAE latents",
        description="Ablating the top 2 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_2: float | None = Field(
        None,
        title="SCR Metric, Top 2 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 2 SAE latents",
    )
    scr_dir2_threshold_2: float | None = Field(
        None,
        title="SCR Dir 2, Top 2 SAE latents",
        description="Ablating the top 2 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_5: float | None = Field(
        None,
        title="SCR Dir 1, Top 5 SAE latents",
        description="Ablating the top 5 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_5: float | None = Field(
        None,
        title="SCR Metric, Top 5 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 5 SAE latents",
    )
    scr_dir2_threshold_5: float | None = Field(
        None,
        title="SCR Dir 2, Top 5 SAE latents",
        description="Ablating the top 5 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_10: float | None = Field(
        None,
        title="SCR Dir 1, Top 10 SAE latents",
        description="Ablating the top 10 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_10: float | None = Field(
        None,
        title="SCR Metric, Top 10 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 10 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    scr_dir2_threshold_10: float | None = Field(
        None,
        title="SCR Dir 2, Top 10 SAE latents",
        description="Ablating the top 10 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_20: float | None = Field(
        None,
        title="SCR Dir 1, Top 20 SAE latents",
        description="Ablating the top 20 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_20: float | None = Field(
        None,
        title="SCR Metric, Top 20 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 20 SAE latents",
    )
    scr_dir2_threshold_20: float | None = Field(
        None,
        title="SCR Dir 2, Top 20 SAE latents",
        description="Ablating the top 20 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_50: float | None = Field(
        None,
        title="SCR Dir 1, Top 50 SAE latents",
        description="Ablating the top 50 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_50: float | None = Field(
        None,
        title="SCR Metric, Top 50 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 50 SAE latents",
    )
    scr_dir2_threshold_50: float | None = Field(
        None,
        title="SCR Dir 2, Top 50 SAE latents",
        description="Ablating the top 50 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_100: float | None = Field(
        None,
        title="SCR Dir 1, Top 100 SAE latents",
        description="Ablating the top 100 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_100: float | None = Field(
        None,
        title="SCR Metric, Top 100 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 100 SAE latents",
    )
    scr_dir2_threshold_100: float | None = Field(
        None,
        title="SCR Dir 2, Top 100 SAE latents",
        description="Ablating the top 100 profession latents to increase gender accuracy",
    )
    scr_dir1_threshold_500: float | None = Field(
        None,
        title="SCR Dir 1, Top 500 SAE latents",
        description="Ablating the top 500 gender latents to increase profession accuracy",
    )
    scr_metric_threshold_500: float | None = Field(
        None,
        title="SCR Metric, Top 500 SAE latents",
        description="SCR Metric (selecting dir1 if inital profession accuracy is lower than initial gender accuracy, else dir2) ablating the top 500 SAE latents",
    )
    scr_dir2_threshold_500: float | None = Field(
        None,
        title="SCR Dir 2, Top 500 SAE latents",
        description="Ablating the top 500 profession latents to increase gender accuracy",
    )


@dataclass(config=ConfigDict(title="SCR"))
class ScrEvalOutput(
    BaseEvalOutput[ScrAndTppEvalConfig, ScrMetricCategories, ScrResultDetail]
):
    """
    The Spurious Correlation Removal (SCR) evaluation ablates SAE latents to shift the bias of a biased linear probe. The methodology is from `Evaluating Sparse Autoencoders on Targeted Concept Removal Tasks`.
    """

    eval_config: ScrAndTppEvalConfig
    eval_id: str
    datetime_epoch_millis: int
    eval_result_metrics: ScrMetricCategories
    eval_result_details: list[ScrResultDetail] = Field(
        default_factory=list,
        title="Per-Dataset Spurious Correlation Removal (SCR) Results",
        description="Each object is a stat on the SCR results for a single dataset.",
    )
    eval_type_id: str = Field(
        default=EVAL_TYPE_ID_SCR,
        title="Eval Type ID",
        description="The type of the evaluation",
    )


# ========= TPP Output


@dataclass
class TppMetrics(BaseMetrics):
    tpp_threshold_2_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 2 SAE latents",
        description="TPP metric when ablating the top 2 SAE latents",
    )
    tpp_threshold_2_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 2 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 2 SAE latents",
    )
    tpp_threshold_2_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 2 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 2 SAE latents",
    )
    tpp_threshold_5_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 5 SAE latents",
        description="TPP metric when ablating the top 5 SAE latents",
    )
    tpp_threshold_5_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 5 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 5 SAE latents",
    )
    tpp_threshold_5_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 5 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 5 SAE latents",
    )
    tpp_threshold_10_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 10 SAE latents",
        description="TPP metric when ablating the top 10 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    tpp_threshold_10_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 10 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 10 SAE latents",
    )
    tpp_threshold_10_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 10 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 10 SAE latents",
    )
    tpp_threshold_20_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 20 SAE latents",
        description="TPP metric when ablating the top 20 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    tpp_threshold_20_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 20 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 20 SAE latents",
    )
    tpp_threshold_20_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 20 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 20 SAE latents",
    )
    tpp_threshold_50_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 50 SAE latents",
        description="TPP metric when ablating the top 50 SAE latents",
    )
    tpp_threshold_50_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 50 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 50 SAE latents",
    )
    tpp_threshold_50_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 50 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 50 SAE latents",
    )
    tpp_threshold_100_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 100 SAE latents",
        description="TPP metric when ablating the top 100 SAE latents",
    )
    tpp_threshold_100_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 100 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 100 SAE latents",
    )
    tpp_threshold_100_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 100 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 100 SAE latents",
    )
    tpp_threshold_500_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 500 SAE latents",
        description="TPP metric when ablating the top 500 SAE latents",
    )
    tpp_threshold_500_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 500 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 500 SAE latents",
    )
    tpp_threshold_500_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 500 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 500 SAE latents",
    )


@dataclass
class TppMetricCategories(BaseMetricCategories):
    tpp_metrics: TppMetrics = Field(
        title="TPP Metrics",
        description="Targeted Probe Perturbation (TPP) results",
        json_schema_extra=DEFAULT_DISPLAY,
    )


@dataclass
class TppResultDetail(BaseResultDetail):
    dataset_name: str = Field(title="Dataset Name", description="")

    tpp_threshold_2_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 2 SAE latents",
        description="TPP metric when ablating the top 2 SAE latents",
    )
    tpp_threshold_2_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 2 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 2 SAE latents",
    )
    tpp_threshold_2_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 2 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 2 SAE latents",
    )
    tpp_threshold_5_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 5 SAE latents",
        description="TPP metric when ablating the top 5 SAE latents",
    )
    tpp_threshold_5_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 5 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 5 SAE latents",
    )
    tpp_threshold_5_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 5 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 5 SAE latents",
    )
    tpp_threshold_10_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 10 SAE latents",
        description="TPP metric when ablating the top 10 SAE latents",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    tpp_threshold_10_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 10 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 10 SAE latents",
    )
    tpp_threshold_10_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 10 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 10 SAE latents",
    )
    tpp_threshold_20_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 20 SAE latents",
        description="TPP metric when ablating the top 20 SAE latents",
    )
    tpp_threshold_20_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 20 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 20 SAE latents",
    )
    tpp_threshold_20_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 20 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 20 SAE latents",
    )
    tpp_threshold_50_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 50 SAE latents",
        description="TPP metric when ablating the top 50 SAE latents",
    )
    tpp_threshold_50_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 50 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 50 SAE latents",
    )
    tpp_threshold_50_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 50 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 50 SAE latents",
    )
    tpp_threshold_100_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 100 SAE latents",
        description="TPP metric when ablating the top 100 SAE latents",
    )
    tpp_threshold_100_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 100 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 100 SAE latents",
    )
    tpp_threshold_100_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 100 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 100 SAE latents",
    )
    tpp_threshold_500_total_metric: float | None = Field(
        None,
        title="TPP Metric, Top 500 SAE latents",
        description="TPP metric when ablating the top 500 SAE latents",
    )
    tpp_threshold_500_intended_diff_only: float | None = Field(
        None,
        title="TPP Intended Class, Top 500 SAE latents",
        description="TPP decrease to the intended class only when ablating the top 500 SAE latents",
    )
    tpp_threshold_500_unintended_diff_only: float | None = Field(
        None,
        title="TPP Unintended Class, Top 500 SAE latents",
        description="TPP decrease to all unintended classes when ablating the top 500 SAE latents",
    )


@dataclass(config=ConfigDict(title="TPP"))
class TppEvalOutput(
    BaseEvalOutput[ScrAndTppEvalConfig, TppMetricCategories, TppResultDetail]
):
    """
    The Targeted Probe Pertubation (TPP) evaluation ablates a set of SAE latents to damage a single targeted linear probe. The methodology is from `Evaluating Sparse Autoencoders on Targeted Concept Removal Tasks`.
    """

    eval_config: ScrAndTppEvalConfig
    eval_id: str
    datetime_epoch_millis: int
    eval_result_metrics: TppMetricCategories
    eval_result_details: list[TppResultDetail] = Field(
        default_factory=list,
        title="Per-Dataset TPP Results",
        description="Each object is a stat on the TPP results for a single dataset.",
    )
    eval_type_id: str = Field(
        default=EVAL_TYPE_ID_TPP,
        title="Eval Type ID",
        description="The type of the evaluation",
    )



import sae_lens
import torch
import torch.nn as nn
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizerBase,
)

import sae_bench.sae_bench_utils.activation_collection as activation_collection
from sae_bench.evals.ravel.eval_config import RAVELEvalConfig


class MDAS(nn.Module):
    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        config: RAVELEvalConfig,
        sae: sae_lens.SAE,  # Kept for API compatibility
    ):
        super().__init__()

        self.model = model
        self.tokenizer = tokenizer
        self.sae = sae  # Kept for API compatibility
        self.layer_intervened = sae.cfg.hook_layer

        # Get the hidden dimension from the model
        hidden_dim = model.config.hidden_size

        # Initialize a square transformation matrix
        # Using Xavier/Glorot initialization for better training dynamics
        # self.transform_matrix = torch.nn.Parameter(
        #     torch.nn.init.xavier_uniform_(
        #         torch.zeros(
        #             hidden_dim, hidden_dim, device=model.device, dtype=model.dtype
        #         )
        #     ),
        #     requires_grad=True,
        # )

        # Add identity initialization option (can be uncommented if needed)
        self.transform_matrix = torch.nn.Parameter(
            torch.eye(hidden_dim, device=model.device, dtype=torch.float32),
            requires_grad=True,
        )
        self.binary_mask = torch.nn.Parameter(
            torch.zeros(hidden_dim, device=model.device, dtype=torch.float32),
            requires_grad=True,
        )

        self.batch_size = config.llm_batch_size
        self.device = model.device
        self.temperature = 1e-2  # Kept for API compatibility

    def create_intervention_hook(
        self,
        source_rep_BD: torch.Tensor,
        base_pos_B: torch.Tensor,
        training_mode: bool = False,
    ):
        def intervention_hook(module, inputs, outputs):
            if isinstance(outputs, tuple):
                resid_BLD = outputs[0]
                rest = outputs[1:]
            else:
                raise ValueError("Unexpected output shape")

            if resid_BLD.shape[1] == 1:
                # This means we are generating with the KV cache and the intervention has already been applied
                return outputs

            # Get the base activations at the target position
            resid_BD = resid_BLD[list(range(resid_BLD.shape[0])), base_pos_B, :]

            # Apply the transformation matrix directly to the source representation
            # This gives us much more flexibility than the binary mask
            rotated_source_BD = torch.matmul(
                source_rep_BD.to(dtype=torch.float32), self.transform_matrix
            )
            rotated_resid_BD = torch.matmul(
                resid_BD.to(dtype=torch.float32), self.transform_matrix
            )

            # Use true binary mask in eval mode, sigmoid in training mode
            if not training_mode:
                mask_values_D = (self.binary_mask > 0).to(dtype=self.binary_mask.dtype)
            else:
                mask_values_D = torch.sigmoid(self.binary_mask / self.temperature)

            # use this to hardcode the mask
            # mask_values_D = torch.zeros_like(binary_mask_D)
            # mask_values_D[:50] = 1

            modified_resid_BD = (
                1 - mask_values_D
            ) * rotated_resid_BD + mask_values_D * rotated_source_BD

            modified_resid_BD = torch.matmul(modified_resid_BD, self.transform_matrix.T)

            # Replace the base activations with the transformed source activations
            resid_BLD[list(range(resid_BLD.shape[0])), base_pos_B, :] = (
                modified_resid_BD.to(dtype=resid_BLD.dtype)
            )

            return (resid_BLD, *rest)

        return intervention_hook

    def forward(
        self,
        base_encoding_BL,
        source_encoding_BL,
        base_pos_B,
        source_pos_B,
        training_mode: bool = False,  # Kept for API compatibility
    ):
        with torch.no_grad():
            # Get source representation
            source_rep = activation_collection.get_layer_activations(
                self.model, self.layer_intervened, source_encoding_BL, source_pos_B
            )

        intervention_hook = self.create_intervention_hook(
            source_rep,
            base_pos_B,
            training_mode,
        )

        handle = activation_collection.get_module(
            self.model, self.layer_intervened
        ).register_forward_hook(intervention_hook)

        logits_BV = self.model(
            input_ids=base_encoding_BL["input_ids"].to(self.model.device),
            attention_mask=base_encoding_BL.get("attention_mask", None),
        ).logits[:, -1, :]

        handle.remove()

        predicted_B = logits_BV.argmax(dim=-1)

        # Format outputs
        predicted_text = []
        for i in range(logits_BV.shape[0]):
            predicted_text.append(self.tokenizer.decode(predicted_B[i]))

        return logits_BV, predicted_text



import subprocess
import uuid
from importlib.metadata import version


def get_eval_uuid():
    return str(uuid.uuid4())


def get_sae_lens_version():
    try:
        return version("sae_lens")
    except Exception:
        return "Unknown"


def get_sae_bench_version():
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode("ascii")
            .strip()
        )
    except Exception:
        return "Unknown"



import random

import numpy as np
import torch
from tqdm import tqdm
from transformers import PreTrainedModel, PreTrainedTokenizerBase

import sae_bench.evals.ravel.mdbm as mdbm
import sae_bench.sae_bench_utils.activation_collection as activation_collection
from sae_bench.evals.ravel.instance import (
    Prompt,
    RAVELFilteredDataset,
    evaluate_completion,
)


def get_different_attribute_prompt(
    base_prompt: Prompt, source_prompts: list[Prompt]
) -> Prompt:
    """
    Select a random prompt from source_prompts that has a different attribute_label
    than the base_prompt.
    """
    different_prompts = [
        p for p in source_prompts if p.attribute_label != base_prompt.attribute_label
    ]
    if not different_prompts:
        raise ValueError(
            f"No prompts with different attribute label found for {base_prompt.attribute_label}"
        )
    return random.choice(different_prompts)


def sample_prompts_by_attribute(
    dataset: RAVELFilteredDataset, attribute: str, n_samples: int
):
    all_prompts = dataset.get_prompts_by_attribute(attribute)
    if len(all_prompts) < n_samples:
        print(
            f"Warning: Not enough prompts for attribute {attribute} for intervention. Returning {len(all_prompts)} instead of {n_samples} prompts."
        )
        return all_prompts, all_prompts

    selected_prompts = random.sample(all_prompts, n_samples)
    return all_prompts, selected_prompts


def get_prompt_pairs(
    dataset: RAVELFilteredDataset,
    base_attribute: str,
    source_attribute: str,
    n_interventions: int,
):
    """
    Selects pairs of base_prompts and source_prompts for the cause and isolation evaluations.
    Base_prompts always contain attribute A templates.
    The cause evaluation requires source_prompts from attribute A templates, attribute values in base and source should differ.
    The isolation evaluation requires source_prompts from attribute B templates.
    """
    all_base_prompts, base_prompts = sample_prompts_by_attribute(
        dataset, base_attribute, n_interventions
    )

    if base_attribute != source_attribute:
        _, source_prompts = sample_prompts_by_attribute(
            dataset, source_attribute, n_interventions
        )
    else:
        all_source_prompts = all_base_prompts
        source_prompts = []
        for p in base_prompts:
            source_prompts.append(get_different_attribute_prompt(p, all_source_prompts))

    min_length = min(len(base_prompts), len(source_prompts))
    return base_prompts[:min_length], source_prompts[:min_length]


@torch.no_grad()
def generate_batched_interventions(
    model: PreTrainedModel,
    mdbm: mdbm.MDBM,
    tokenizer: PreTrainedTokenizerBase,
    val_loader: torch.utils.data.DataLoader,
    max_new_tokens: int = 8,
) -> tuple[float, float]:
    iso_scores = []
    cause_scores = []
    for batch in tqdm(val_loader, desc="Generating with interventions"):
        (
            base_encoding_BL,
            source_encoding_BL,
            base_pos_B,
            source_pos_B,
            base_pred_B,
            source_pred_B,
            base_text_str,
            base_label_str,
        ) = batch

        # Get source representation
        source_rep_BD = activation_collection.get_layer_activations(
            model, mdbm.layer_intervened, source_encoding_BL, source_pos_B
        )

        intervention_hook = mdbm.create_intervention_hook(
            source_rep_BD,
            base_pos_B,
            training_mode=False,
        )

        handle = activation_collection.get_module(
            model, mdbm.layer_intervened
        ).register_forward_hook(intervention_hook)

        # Generate using huggingface model
        output_ids = model.generate(
            input_ids=base_encoding_BL["input_ids"].to(model.device),
            attention_mask=base_encoding_BL.get("attention_mask", None).to(
                model.device
            ),
            max_new_tokens=max_new_tokens,
            do_sample=False,  # greedy decoding for reproducibility
        )  # type: ignore

        handle.remove()
        generated_ids = output_ids[:, -max_new_tokens:]
        generated_strings = tokenizer.batch_decode(generated_ids)

        for base_text, base_label, generated_string, base_pred, source_pred in zip(
            base_text_str, base_label_str, generated_strings, base_pred_B, source_pred_B
        ):
            if base_pred == source_pred:
                iso_scores.append(
                    evaluate_completion(base_text, base_label, generated_string)
                )
            else:
                cause_scores.append(
                    evaluate_completion(base_text, base_label, generated_string)
                )

    iso_score = float(np.mean(iso_scores))
    cause_score = float(np.mean(cause_scores))
    return iso_score, cause_score



import random
from collections import defaultdict

import einops
import pandas as pd
import torch
from datasets import load_dataset
from tqdm import tqdm
from transformers import AutoTokenizer

import sae_bench.sae_bench_utils.dataset_info as dataset_info


def get_dataset_list_of_strs(
    dataset_name: str, column_name: str, min_row_chars: int, total_chars: int
) -> list[str]:
    dataset = load_dataset(dataset_name, split="train", streaming=True)

    total_chars_so_far = 0
    result = []

    for row in dataset:
        if len(row[column_name]) > min_row_chars:  # type: ignore
            result.append(row[column_name])  # type: ignore
            total_chars_so_far += len(row[column_name])  # type: ignore
            if total_chars_so_far > total_chars:
                break

    return result


def load_and_tokenize_dataset(
    dataset_name: str,
    ctx_len: int,
    num_tokens: int,
    tokenizer: AutoTokenizer,
    column_name: str = "text",
    add_bos: bool = True,
) -> torch.Tensor:
    dataset = get_dataset_list_of_strs(dataset_name, column_name, 100, num_tokens * 5)

    tokens = tokenize_and_concat_dataset(
        tokenizer, dataset, ctx_len, add_bos=add_bos, max_tokens=num_tokens
    )

    assert (tokens.shape[0] * tokens.shape[1]) > num_tokens

    return tokens


def tokenize_and_concat_dataset(
    tokenizer: AutoTokenizer,
    dataset: list[str],
    seq_len: int,
    add_bos: bool = True,
    max_tokens: int | None = None,
) -> torch.Tensor:
    full_text = tokenizer.eos_token.join(dataset)  # type: ignore

    # divide into chunks to speed up tokenization
    num_chunks = 20
    chunk_length = (len(full_text) - 1) // num_chunks + 1
    chunks = [
        full_text[i * chunk_length : (i + 1) * chunk_length] for i in range(num_chunks)
    ]
    tokens = tokenizer(chunks, return_tensors="pt", padding=True)["input_ids"].flatten()  # type: ignore

    # remove pad token
    tokens = tokens[tokens != tokenizer.pad_token_id]  # type: ignore

    if max_tokens is not None:
        tokens = tokens[: max_tokens + seq_len + 1]

    num_tokens = len(tokens)
    num_batches = num_tokens // seq_len

    # drop last batch if not full
    tokens = tokens[: num_batches * seq_len]
    tokens = einops.rearrange(
        tokens, "(batch seq) -> batch seq", batch=num_batches, seq=seq_len
    )

    if add_bos:
        tokens[:, 0] = tokenizer.bos_token_id  # type: ignore
    return tokens


def gather_dataset_from_df(
    df: pd.DataFrame,
    chosen_classes: list[str],
    min_samples_per_category: int,
    label_key: str,
    text_key: str,
    random_seed: int,
) -> dict[str, list[str]]:
    random.seed(random_seed)

    data = {}

    for chosen_class in chosen_classes:
        class_df = df[df[label_key] == chosen_class]

        sampled_texts = (
            class_df[text_key]
            .sample(n=min_samples_per_category, random_state=random_seed)  # type: ignore
            .tolist()
        )
        assert len(sampled_texts) == min_samples_per_category

        data[str(chosen_class)] = sampled_texts

    return data


def get_ag_news_dataset(
    dataset_name: str,
    chosen_classes: list[str],
    train_set_size: int,
    test_set_size: int,
    random_seed: int,
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    random.seed(random_seed)

    dataset = load_dataset(dataset_name, streaming=False)
    train_df = pd.DataFrame(dataset["train"])  # type: ignore
    test_df = pd.DataFrame(dataset["test"])  # type: ignore

    # It's a binary classification task, so we need to halve the train and test sizes
    train_size = train_set_size // 2
    test_size = test_set_size // 2

    # convert str to int, as labels are stored as ints
    chosen_classes = [int(chosen_class) for chosen_class in chosen_classes]  # type: ignore

    train_data = gather_dataset_from_df(
        train_df, chosen_classes, train_size, "label", "text", random_seed
    )
    test_data = gather_dataset_from_df(
        test_df, chosen_classes, test_size, "label", "text", random_seed
    )

    return train_data, test_data


def get_europarl_dataset(
    dataset_name: str,
    chosen_languages: list[str],
    train_size: int,
    test_size: int,
    random_seed: int,
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    random.seed(random_seed)
    label_key = "translation"
    language_pairs = {
        "en": "en-fr",
        "fr": "fr-it",
        "de": "de-en",
        "es": "es-fr",
        "nl": "nl-pt",
    }

    # It's a binary classification task, so we need to halve the train and test sizes
    train_size = train_size // 2
    test_size = test_size // 2

    samples_per_language = train_size + test_size

    samples_by_language = defaultdict(list)

    print(f"Loading dataset {dataset_name}, this usually takes ~10 seconds")

    for language, language_pair in language_pairs.items():
        # Filter out languages that are not in the dataset
        dataset = load_dataset(
            dataset_name,
            language_pair,
            streaming=True,
            split="train",
        )

        # Collect samples for each language
        for sample in dataset:
            # Extract the text in the target language
            text = sample[label_key][language]  # type: ignore
            samples_by_language[language].append(text)

            # Check if we have enough samples for all languages
            if len(samples_by_language[language]) > samples_per_language:
                break

    # Split samples into train and test sets
    train_samples = {}
    test_samples = {}

    for language in chosen_languages:
        lang_samples = samples_by_language[language]

        random.shuffle(lang_samples)
        train_samples[language] = lang_samples[:train_size]
        test_samples[language] = lang_samples[train_size : train_size + test_size]
        assert len(train_samples[language]) == train_size
        assert len(test_samples[language]) == test_size

    return train_samples, test_samples


def get_github_code_dataset(
    dataset_name: str,
    chosen_classes: list[str],
    train_size: int,
    test_size: int,
    random_seed: int,
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    """Following the Neurons in a Haystack paper, we skip the first 50 tokens of each code snippet to avoid the license header.
    We use characters instead of tokens to avoid tokenization issues."""
    tokens_to_skip = 50
    ctx_len = 128
    chars_per_token = 3
    ctx_len_chars = ctx_len * chars_per_token
    chars_to_skip = tokens_to_skip * chars_per_token

    random.seed(random_seed)
    label_key = "language"

    # It's a binary classification task, so we need to halve the train and test sizes
    train_size = train_size // 2
    test_size = test_size // 2

    print(f"Loading dataset {dataset_name}, this usually takes ~30 seconds")

    # Filter out languages that are not in the dataset
    dataset = load_dataset(
        dataset_name,
        streaming=True,
        split="train",
        trust_remote_code=True,
        languages=chosen_classes,
    )

    total_size = train_size + test_size

    all_samples = defaultdict(list)

    # Collect samples for each language
    for sample in dataset:
        if sample[label_key] in chosen_classes:  # type: ignore
            code = sample["code"]  # type: ignore

            # In "Neurons in a Haystack", the authors skipped the first 50 tokens to avoid the license header
            # This is using characters so it's tokenizer agnostic
            if len(code) > (ctx_len_chars + chars_to_skip):
                code = code[chars_to_skip:]
                all_samples[sample[label_key]].append(code)  # type: ignore

            # Check if we have collected enough samples for all languages
            if all(len(all_samples[lang]) > total_size for lang in chosen_classes):
                break

    # Split samples into train and test sets
    train_samples = {}
    test_samples = {}

    for lang in chosen_classes:
        lang_samples = all_samples[lang]

        random.shuffle(lang_samples)
        train_samples[lang] = lang_samples[:train_size]
        test_samples[lang] = lang_samples[train_size : train_size + test_size]
        assert len(train_samples[lang]) == train_size
        assert len(test_samples[lang]) == test_size

    return train_samples, test_samples


def get_balanced_dataset(
    df: pd.DataFrame,
    dataset_name: str,
    min_samples_per_quadrant: int,
    random_seed: int,
) -> dict[str, list[str]]:
    """This function is used for the amazon reviews dataset and the bias_in_bios dataset, which have two columns.

    Returns a balanced dataset as a dictionary, where each key corresponds to a unique value
    in one column, and each value is a list of text entries balanced across categories
    in the other column.

    Examples: For the 'bias_in_bios' dataset where `column1` is 'Profession' and `column2` is 'Gender':
        - If `balance_by_column1` is `True`:
            - Balances bios for each profession by gender.
            - Returns a dict with professions as keys and lists of bios as values.
    """

    text_column_name = dataset_info.dataset_metadata[dataset_name]["text_column_name"]
    column1_name = dataset_info.dataset_metadata[dataset_name]["column1_name"]
    column2_name = dataset_info.dataset_metadata[dataset_name]["column2_name"]

    balanced_data = {}

    for profession in tqdm(df[column1_name].unique()):
        prof_df = df[df[column1_name] == profession]
        unique_groups = prof_df[column2_name].unique()  # type: ignore
        min_count = prof_df[column2_name].value_counts().min()  # type: ignore

        if len(unique_groups) < 2:
            continue  # Skip professions with less than two groups

        if min_count < min_samples_per_quadrant:
            continue

        sampled_texts = []
        for _, group_df in prof_df.groupby(column2_name):
            sampled_group = group_df.sample(
                n=min_samples_per_quadrant, random_state=random_seed
            )
            sampled_texts.extend(sampled_group[text_column_name].tolist())

        balanced_data[str(profession)] = sampled_texts

        assert len(balanced_data[str(profession)]) == min_samples_per_quadrant * 2

    return balanced_data


def get_bias_in_bios_or_amazon_product_dataset(
    dataset_name: str, train_set_size: int, test_set_size: int, random_seed: int
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    dataset_name = dataset_name.split("_class_set")[0]

    dataset = load_dataset(dataset_name)
    train_df = pd.DataFrame(dataset["train"])  # type: ignore
    test_df = pd.DataFrame(dataset["test"])  # type: ignore

    # 4 is because male / female split for each profession, 2 quadrants per profession, 2 professions for binary task
    minimum_train_samples_per_quadrant = train_set_size // 4
    minimum_test_samples_per_quadrant = test_set_size // 4

    train_data = get_balanced_dataset(
        train_df, dataset_name, minimum_train_samples_per_quadrant, random_seed
    )
    test_data = get_balanced_dataset(
        test_df, dataset_name, minimum_test_samples_per_quadrant, random_seed
    )

    return train_data, test_data


def get_amazon_sentiment_dataset(
    dataset_name: str, train_set_size: int, test_set_size: int, random_seed: int
) -> tuple[pd.DataFrame, pd.DataFrame]:
    dataset_name = dataset_name.split("_sentiment")[0]
    dataset = load_dataset(dataset_name)
    train_df = pd.DataFrame(dataset["train"])  # type: ignore
    test_df = pd.DataFrame(dataset["test"])  # type: ignore

    minimum_train_samples_per_category = train_set_size // 2
    minimum_test_samples_per_category = test_set_size // 2

    train_data = get_balanced_amazon_sentiment_dataset(
        train_df, minimum_train_samples_per_category, random_seed
    )
    test_data = get_balanced_amazon_sentiment_dataset(
        test_df, minimum_test_samples_per_category, random_seed
    )

    return train_data, test_data  # type: ignore


def get_balanced_amazon_sentiment_dataset(
    df: pd.DataFrame,
    min_samples_per_category: int,
    random_seed: int,
) -> dict[str, list[str]]:
    text_column_name = "text"
    column2_name = "rating"

    balanced_data = {}

    unique_ratings = df[column2_name].unique()

    for rating in unique_ratings:
        # Filter dataframe for current rating
        df_rating = df[df[column2_name] == rating]

        sampled_texts = (
            df_rating[text_column_name]
            .sample(n=min_samples_per_category, random_state=random_seed)  # type: ignore
            .tolist()
        )
        assert len(sampled_texts) == min_samples_per_category

        balanced_data[str(rating)] = sampled_texts

    return balanced_data


def ensure_shared_keys(train_data: dict, test_data: dict) -> tuple[dict, dict]:
    # Find keys that are in test but not in train
    test_only_keys = set(test_data.keys()) - set(train_data.keys())

    # Find keys that are in train but not in test
    train_only_keys = set(train_data.keys()) - set(test_data.keys())

    # Remove keys from test that are not in train
    for key in test_only_keys:
        print(f"Removing {key} from test set")
        del test_data[key]

    # Remove keys from train that are not in test
    for key in train_only_keys:
        print(f"Removing {key} from train set")
        del train_data[key]

    return train_data, test_data


def get_multi_label_train_test_data(
    dataset_name: str,
    train_set_size: int,
    test_set_size: int,
    random_seed: int,
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    """Returns a dict of [class_name, list[str]]"""

    if (
        "bias_in_bios" in dataset_name
        or "canrager/amazon_reviews_mcauley_1and5" == dataset_name
    ):
        train_data, test_data = get_bias_in_bios_or_amazon_product_dataset(
            dataset_name, train_set_size, test_set_size, random_seed
        )
    elif dataset_name == "canrager/amazon_reviews_mcauley_1and5_sentiment":
        train_data, test_data = get_amazon_sentiment_dataset(
            dataset_name, train_set_size, test_set_size, random_seed
        )
    elif dataset_name == "codeparrot/github-code":
        train_data, test_data = get_github_code_dataset(
            dataset_name,
            dataset_info.chosen_classes_per_dataset[dataset_name],
            train_set_size,
            test_set_size,
            random_seed,
        )
    elif dataset_name == "fancyzhx/ag_news":
        train_data, test_data = get_ag_news_dataset(
            dataset_name,
            dataset_info.chosen_classes_per_dataset[dataset_name],
            train_set_size,
            test_set_size,
            random_seed,
        )
    elif dataset_name == "Helsinki-NLP/europarl":
        train_data, test_data = get_europarl_dataset(
            dataset_name,
            dataset_info.chosen_classes_per_dataset[dataset_name],
            train_set_size,
            test_set_size,
            random_seed,
        )
    else:
        raise ValueError(f"Dataset {dataset_name} not supported")

    train_data, test_data = ensure_shared_keys(train_data, test_data)  # type: ignore

    return train_data, test_data


def tokenize_data_dictionary(
    data: dict[str, list[str]], tokenizer: AutoTokenizer, max_length: int, device: str
) -> dict[str, dict]:
    tokenized_data = {}
    for key, texts in tqdm(data.items(), desc="Tokenizing data"):
        # .data so we have a dict, not a BatchEncoding
        tokenized_data[key] = (
            tokenizer(
                texts,
                padding="max_length",
                truncation=True,
                max_length=max_length,
                return_tensors="pt",
            )  # type: ignore
            .to(device)
            .data
        )
    return tokenized_data


def filter_dataset(
    data: dict[str, list[str]], chosen_class_indices: list[str]
) -> dict[str, list[str]]:
    filtered_data = {}
    for class_name in chosen_class_indices:
        filtered_data[class_name] = data[class_name]
    return filtered_data



import json
import os

import torch
from huggingface_hub import snapshot_download
from tqdm import tqdm

import sae_bench.custom_saes.base_sae as base_sae
import sae_bench.custom_saes.batch_topk_sae as batch_topk_sae
import sae_bench.custom_saes.gated_sae as gated_sae
import sae_bench.custom_saes.jumprelu_sae as jumprelu_sae
import sae_bench.custom_saes.relu_sae as relu_sae
import sae_bench.custom_saes.topk_sae as topk_sae
import sae_bench.evals.absorption.main as absorption
import sae_bench.evals.autointerp.main as autointerp
import sae_bench.evals.core.main as core
import sae_bench.evals.ravel.main as ravel
import sae_bench.evals.scr_and_tpp.main as scr_and_tpp
import sae_bench.evals.sparse_probing.main as sparse_probing
import sae_bench.evals.unlearning.main as unlearning
import sae_bench.sae_bench_utils.general_utils as general_utils

MODEL_CONFIGS = {
    "pythia-70m-deduped": {
        "batch_size": 512,
        "dtype": "float32",
        "layers": [3, 4],
        "d_model": 512,
    },
    "pythia-160m-deduped": {
        "batch_size": 256,
        "dtype": "float32",
        "layers": [8],
        "d_model": 768,
    },
    "gemma-2-2b": {
        "batch_size": 32,
        "dtype": "bfloat16",
        "layers": [5, 12, 19],
        "d_model": 2304,
    },
}

output_folders = {
    "absorption": "eval_results/absorption",
    "autointerp": "eval_results/autointerp",
    "core": "eval_results/core",
    "scr": "eval_results/scr",
    "tpp": "eval_results/tpp",
    "sparse_probing": "eval_results/sparse_probing",
    "unlearning": "eval_results/unlearning",
    "ravel": "eval_results/ravel",
}


TRAINER_LOADERS = {
    "MatryoshkaBatchTopKTrainer": batch_topk_sae.load_dictionary_learning_matryoshka_batch_topk_sae,
    "BatchTopKTrainer": batch_topk_sae.load_dictionary_learning_batch_topk_sae,
    "TopKTrainer": topk_sae.load_dictionary_learning_topk_sae,
    "StandardTrainerAprilUpdate": relu_sae.load_dictionary_learning_relu_sae,
    "StandardTrainer": relu_sae.load_dictionary_learning_relu_sae,
    "PAnnealTrainer": relu_sae.load_dictionary_learning_relu_sae,
    "JumpReluTrainer": jumprelu_sae.load_dictionary_learning_jump_relu_sae,
    "GatedSAETrainer": gated_sae.load_dictionary_learning_gated_sae,
}


def get_all_hf_repo_autoencoders(
    repo_id: str, download_location: str = "downloaded_saes"
) -> list[str]:
    download_location = os.path.join(download_location, repo_id.replace("/", "_"))
    config_dir = snapshot_download(
        repo_id,
        allow_patterns=["*config.json"],
        local_dir=download_location,
        force_download=False,
    )

    config_locations = []

    for root, _, files in os.walk(config_dir):
        for file in files:
            if file == "config.json":
                config_locations.append(os.path.join(root, file))

    repo_locations = []

    for config in config_locations:
        repo_location = config.split(f"{download_location}/")[1].split("/config.json")[
            0
        ]
        repo_locations.append(repo_location)

    return repo_locations


def load_dictionary_learning_sae(
    repo_id: str,
    location: str,
    model_name,
    device: str,
    dtype: torch.dtype,
    layer: int | None = None,
    download_location: str = "downloaded_saes",
) -> base_sae.BaseSAE:
    download_location = os.path.join(download_location, repo_id.replace("/", "_"))

    config_file = f"{download_location}/{location}/config.json"

    with open(config_file) as f:
        config = json.load(f)

    trainer_class = config["trainer"]["trainer_class"]

    location = f"{location}/ae.pt"

    sae = TRAINER_LOADERS[trainer_class](
        repo_id=repo_id,
        filename=location,
        layer=layer,
        model_name=model_name,
        device=device,
        dtype=dtype,
    )
    return sae


def verify_saes_load(
    repo_id: str,
    sae_locations: list[str],
    model_name: str,
    device: str,
    dtype: torch.dtype,
):
    """Verify that all SAEs load correctly. Useful to check this before a big evaluation run."""
    for sae_location in sae_locations:
        sae = load_dictionary_learning_sae(
            repo_id=repo_id,
            location=sae_location,
            layer=None,
            model_name=model_name,
            device=device,
            dtype=dtype,
        )
        del sae


def run_evals(
    repo_id: str,
    model_name: str,
    sae_locations: list[str],
    llm_batch_size: int,
    llm_dtype: str,
    device: str,
    eval_types: list[str],
    random_seed: int,
    api_key: str | None = None,
    force_rerun: bool = False,
):
    """Run selected evaluations for the given model and SAEs."""

    if model_name not in MODEL_CONFIGS:
        raise ValueError(f"Unsupported model: {model_name}")

    # Mapping of eval types to their functions and output paths
    eval_runners = {
        "absorption": (
            lambda selected_saes, is_final: absorption.run_eval(
                absorption.AbsorptionEvalConfig(
                    model_name=model_name,
                    random_seed=random_seed,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results/absorption",
                force_rerun,
            )
        ),
        "autointerp": (
            lambda selected_saes, is_final: autointerp.run_eval(
                autointerp.AutoInterpEvalConfig(
                    model_name=model_name,
                    random_seed=random_seed,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                api_key,  # type: ignore
                "eval_results/autointerp",
                force_rerun,
            )
        ),
        # TODO: Do a better job of setting num_batches and batch size
        # The core run_eval() interface isn't well suited for custom SAEs, so we have to do this instead.
        # It isn't ideal, but it works.
        # TODO: Don't hardcode magic numbers
        "core": (
            lambda selected_saes, is_final: core.multiple_evals(
                selected_saes=selected_saes,
                n_eval_reconstruction_batches=200,
                n_eval_sparsity_variance_batches=2000,
                eval_batch_size_prompts=16,
                compute_featurewise_density_statistics=True,
                compute_featurewise_weight_based_metrics=True,
                exclude_special_tokens_from_reconstruction=True,
                dataset="Skylion007/openwebtext",
                context_size=128,
                output_folder="eval_results/core",
                verbose=True,
                dtype=llm_dtype,
                device=device,
            )
        ),
        "ravel": (
            lambda selected_saes, is_final: ravel.run_eval(
                ravel.RAVELEvalConfig(
                    model_name=model_name,
                    random_seed=random_seed,
                    llm_batch_size=llm_batch_size // 4,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results/ravel",
                force_rerun,
            )
        ),
        "scr": (
            lambda selected_saes, is_final: scr_and_tpp.run_eval(
                scr_and_tpp.ScrAndTppEvalConfig(
                    model_name=model_name,
                    random_seed=random_seed,
                    perform_scr=True,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results",  # We add scr or tpp depending on perform_scr
                force_rerun,
                clean_up_activations=is_final,
                save_activations=True,
            )
        ),
        "tpp": (
            lambda selected_saes, is_final: scr_and_tpp.run_eval(
                scr_and_tpp.ScrAndTppEvalConfig(
                    model_name=model_name,
                    random_seed=random_seed,
                    perform_scr=False,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results",  # We add scr or tpp depending on perform_scr
                force_rerun,
                clean_up_activations=is_final,
                save_activations=True,
            )
        ),
        "sparse_probing": (
            lambda selected_saes, is_final: sparse_probing.run_eval(
                sparse_probing.SparseProbingEvalConfig(
                    model_name=model_name,
                    random_seed=random_seed,
                    llm_batch_size=llm_batch_size,
                    llm_dtype=llm_dtype,
                ),
                selected_saes,
                device,
                "eval_results/sparse_probing",
                force_rerun,
                clean_up_activations=is_final,
                save_activations=True,
            )
        ),
        "unlearning": (
            lambda selected_saes, is_final: unlearning.run_eval(
                unlearning.UnlearningEvalConfig(
                    model_name="gemma-2-2b-it",
                    random_seed=random_seed,
                    llm_dtype=llm_dtype,
                    llm_batch_size=llm_batch_size
                    // 8,  # 8x smaller batch size for unlearning due to longer sequences
                ),
                selected_saes,
                device,
                "eval_results/unlearning",
                force_rerun,
            )
        ),
    }

    for eval_type in eval_types:
        if eval_type not in eval_runners:
            raise ValueError(f"Unsupported eval type: {eval_type}")

    verify_saes_load(
        repo_id,
        sae_locations,
        model_name,
        device,
        general_utils.str_to_dtype(llm_dtype),
    )

    # Run selected evaluations
    for eval_type in tqdm(eval_types, desc="Evaluations"):
        if eval_type == "autointerp" and api_key is None:
            print("Skipping autointerp evaluation due to missing API key")
            continue
        if eval_type == "unlearning":
            if not os.path.exists(
                "./sae_bench/evals/unlearning/data/bio-forget-corpus.jsonl"
            ):
                print(
                    "Skipping unlearning evaluation due to missing bio-forget-corpus.jsonl"
                )
                continue

        print(f"\n\n\nRunning {eval_type} evaluation\n\n\n")

        try:
            for i, sae_location in enumerate(sae_locations):
                is_final = False
                if i == len(sae_locations) - 1:
                    is_final = True

                sae = load_dictionary_learning_sae(
                    repo_id=repo_id,
                    location=sae_location,
                    layer=None,
                    model_name=model_name,
                    device=device,
                    dtype=general_utils.str_to_dtype(llm_dtype),
                )
                unique_sae_id = sae_location.replace("/", "_")
                unique_sae_id = f"{repo_id.split('/')[1]}_{unique_sae_id}"
                selected_saes = [(unique_sae_id, sae)]

                os.makedirs(output_folders[eval_type], exist_ok=True)
                eval_runners[eval_type](selected_saes, is_final)

                del sae

        except Exception as e:
            print(f"Error running {eval_type} evaluation: {e}")
            continue


if __name__ == "__main__":
    """
    This will run all evaluations on all selected dictionary_learning SAEs within the specified HuggingFace repos.
    Set the model_name(s) and repo_id(s) in `repos`.
    Also specify the eval types you want to run in `eval_types`.
    You can also specify any keywords to exclude/include in the SAE filenames using `exclude_keywords` and `include_keywords`.
    NOTE: If your model (with associated model_name and batch sizes) is not in the MODEL_CONFIGS dictionary, you will need to add it.
    This relies on each SAE being located in a folder which contains an ae.pt file and a config.json file (which is the default save format in dictionary_learning).
    Running this script as is should run SAE Bench Pythia and Gemma SAEs.
    """
    RANDOM_SEED = 42

    device = general_utils.setup_environment()

    # Select your eval types here.
    # Note: Unlearning is not recommended for models with < 2B parameters and we recommend an instruct tuned model
    # Unlearning will also require requesting permission for the WMDP dataset (see unlearning/README.md)
    # Absorption not recommended for models < 2B parameters
    eval_types = [
        "absorption",
        "core",
        "scr",
        "tpp",
        "sparse_probing",
        "autointerp",
        # "unlearning",
        "ravel",
    ]

    if "autointerp" in eval_types:
        try:
            with open("openai_api_key.txt") as f:
                api_key = f.read().strip()
        except FileNotFoundError:
            raise Exception("Please create openai_api_key.txt with your API key")
    else:
        api_key = None

    if "unlearning" in eval_types:
        if not os.path.exists(
            "./sae_bench/evals/unlearning/data/bio-forget-corpus.jsonl"
        ):
            raise Exception(
                "Please download bio-forget-corpus.jsonl for unlearning evaluation"
            )

    repos = [
        (
            "adamkarvonen/saebench_pythia-160m-deduped_width-2pow14_date-0108",
            "pythia-160m-deduped",
        ),
        ("canrager/saebench_gemma-2-2b_width-2pow14_date-0107", "gemma-2-2b"),
    ]
    exclude_keywords = ["checkpoints"]
    include_keywords = []

    for repo_id, model_name in repos:
        print(f"\n\n\nEvaluating {model_name} with {repo_id}\n\n\n")

        llm_batch_size = MODEL_CONFIGS[model_name]["batch_size"]
        str_dtype = MODEL_CONFIGS[model_name]["dtype"]
        torch_dtype = general_utils.str_to_dtype(str_dtype)

        sae_locations = get_all_hf_repo_autoencoders(repo_id)

        sae_locations = general_utils.filter_keywords(
            sae_locations,
            exclude_keywords=exclude_keywords,
            include_keywords=include_keywords,
        )

        run_evals(
            repo_id=repo_id,
            model_name=model_name,
            sae_locations=sae_locations,
            llm_batch_size=llm_batch_size,
            llm_dtype=str_dtype,
            device=device,
            eval_types=eval_types,
            api_key=api_key,
            random_seed=RANDOM_SEED,
        )



"""
This file will be deleted once the encode method is merged into SAELens (currently it's not possible to encode with only
a slice of the SAE latents).
"""

import torch
from jaxtyping import Float
from sae_lens import SAE


def encode_subset(
    self: SAE, x: torch.Tensor, latents: torch.Tensor | None = None
) -> torch.Tensor:
    """
    Calculate SAE latents from inputs. Includes optional `latents` argument to only calculate a subset.
    """
    # Get the encoding function for this SAE architecture
    encode_fn = {
        "standard": encode_standard,
        "gated": encode_gated,
        "jumprelu": encode_jumprelu,
    }[self.cfg.architecture]

    # If the activation function is topk, we're required to compute all activations before slicing
    return (
        encode_fn(self, x, latents=None)[..., latents]
        if self.cfg.activation_fn_str == "topk"
        else encode_fn(self, x, latents)
    )


def encode_gated(
    self: SAE,
    x: Float[torch.Tensor, "... d_in"],
    latents: torch.Tensor | None = None,
) -> Float[torch.Tensor, "... d_sae"]:
    """
    Computes the latent values of the Sparse Autoencoder (SAE) using a gated architecture. The activation values are
    computed as the product of the masking term & the post-activation function magnitude term:

        1[(x - b_dec) @ W_gate + b_gate > 0] * activation_fn((x - b_dec) @ W_enc + b_enc)

    The `latents` argument allows for the computation of a specific subset of the hidden values. If `latents` is not
    provided, all latent values will be computed.
    """
    latents_tensor = torch.arange(self.cfg.d_sae) if latents is None else latents

    x = x.to(self.dtype)
    x = self.reshape_fn_in(x)
    x = self.hook_sae_input(x)
    x = self.run_time_activation_norm_fn_in(x)
    sae_in = x - self.b_dec * self.cfg.apply_b_dec_to_input

    # Gating path
    gating_pre_activation = (
        sae_in @ self.W_enc[:, latents_tensor] + self.b_gate[latents_tensor]
    )
    active_features = (gating_pre_activation > 0).to(self.dtype)

    # Magnitude path with weight sharing
    magnitude_pre_activation = self.hook_sae_acts_pre(
        sae_in @ (self.W_enc[:, latents_tensor] * self.r_mag[latents_tensor].exp())
        + self.b_mag[latents_tensor]
    )
    feature_magnitudes = self.activation_fn(magnitude_pre_activation)

    feature_acts = self.hook_sae_acts_post(active_features * feature_magnitudes)

    return feature_acts


def encode_jumprelu(
    self: SAE,
    x: Float[torch.Tensor, "... d_in"],
    latents: torch.Tensor | None = None,
) -> Float[torch.Tensor, "... d_sae"]:
    """
    Computes the latent values of the Sparse Autoencoder (SAE) using a gated architecture. The activation values are
    computed as:

        activation_fn((x - b_dec) @ W_enc + b_enc) * 1[(x - b_dec) @ W_enc + b_enc > threshold]

    The `latents` argument allows for the computation of a specific subset of the hidden values. If `latents` is not
    provided, all latent values will be computed.
    """
    latents_tensor = torch.arange(self.cfg.d_sae) if latents is None else latents

    # move x to correct dtype
    x = x.to(self.dtype)

    # handle hook z reshaping if needed.
    x = self.reshape_fn_in(x)  # type: ignore

    # handle run time activation normalization if needed
    x = self.run_time_activation_norm_fn_in(x)

    # apply b_dec_to_input if using that method.
    sae_in = self.hook_sae_input(x - (self.b_dec * self.cfg.apply_b_dec_to_input))

    # "... d_in, d_in d_sae -> ... d_sae",
    hidden_pre = self.hook_sae_acts_pre(
        sae_in @ self.W_enc[:, latents_tensor] + self.b_enc[latents_tensor]
    )

    feature_acts = self.hook_sae_acts_post(
        self.activation_fn(hidden_pre) * (hidden_pre > self.threshold[latents_tensor])
    )

    return feature_acts


def encode_standard(
    self: SAE,
    x: Float[torch.Tensor, "... d_in"],
    latents: torch.Tensor | None = None,
) -> Float[torch.Tensor, "... d_sae"]:
    """
    Computes the latent values of the Sparse Autoencoder (SAE) using a gated architecture. The activation values are
    computed as:

        activation_fn((x - b_dec) @ W_enc + b_enc)

    The `latents` argument allows for the computation of a specific subset of the hidden values. If `latents` is not
    provided, all latent values will be computed.
    """
    latents_tensor = torch.arange(self.cfg.d_sae) if latents is None else latents

    x = x.to(self.dtype)
    x = self.reshape_fn_in(x)
    x = self.hook_sae_input(x)
    x = self.run_time_activation_norm_fn_in(x)

    # apply b_dec_to_input if using that method.
    sae_in = x - (self.b_dec * self.cfg.apply_b_dec_to_input)

    # "... d_in, d_in d_sae -> ... d_sae",
    hidden_pre = self.hook_sae_acts_pre(
        sae_in @ self.W_enc[:, latents_tensor] + self.b_enc[latents_tensor]
    )
    feature_acts = self.hook_sae_acts_post(self.activation_fn(hidden_pre))

    return feature_acts



# TODO: Consolidate all bias in bios utility stuff
# TODO: Only use strings for keys, only use ints when initializing the dictionary datasets

POSITIVE_CLASS_LABEL = 1
NEGATIVE_CLASS_LABEL = 0

# NOTE: These are going to be hardcoded, and won't change even if the underlying dataset or data labels change.
# This is a bit confusing, but IMO male_professor / female_nurse is a bit easier to understand than e.g. class1_pos_class2_pos / class1_neg_class2_neg
PAIRED_CLASS_KEYS = {
    "male / female": "female_data_only",
    "professor / nurse": "nurse_data_only",
    "male_professor / female_nurse": "female_nurse_data_only",
}

profession_dict = {
    "accountant": 0,
    "architect": 1,
    "attorney": 2,
    "chiropractor": 3,
    "comedian": 4,
    "composer": 5,
    "dentist": 6,
    "dietitian": 7,
    "dj": 8,
    "filmmaker": 9,
    "interior_designer": 10,
    "journalist": 11,
    "model": 12,
    "nurse": 13,
    "painter": 14,
    "paralegal": 15,
    "pastor": 16,
    "personal_trainer": 17,
    "photographer": 18,
    "physician": 19,
    "poet": 20,
    "professor": 21,
    "psychologist": 22,
    "rapper": 23,
    "software_engineer": 24,
    "surgeon": 25,
    "teacher": 26,
    "yoga_teacher": 27,
}
profession_int_to_str = {v: k for k, v in profession_dict.items()}

gender_dict = {
    "male": 0,
    "female": 1,
}

# From the original dataset
amazon_category_dict = {
    "All_Beauty": 0,
    "Toys_and_Games": 1,
    "Cell_Phones_and_Accessories": 2,
    "Industrial_and_Scientific": 3,
    "Gift_Cards": 4,
    "Musical_Instruments": 5,
    "Electronics": 6,
    "Handmade_Products": 7,
    "Arts_Crafts_and_Sewing": 8,
    "Baby_Products": 9,
    "Health_and_Household": 10,
    "Office_Products": 11,
    "Digital_Music": 12,
    "Grocery_and_Gourmet_Food": 13,
    "Sports_and_Outdoors": 14,
    "Home_and_Kitchen": 15,
    "Subscription_Boxes": 16,
    "Tools_and_Home_Improvement": 17,
    "Pet_Supplies": 18,
    "Video_Games": 19,
    "Kindle_Store": 20,
    "Clothing_Shoes_and_Jewelry": 21,
    "Patio_Lawn_and_Garden": 22,
    "Unknown": 23,
    "Books": 24,
    "Automotive": 25,
    "CDs_and_Vinyl": 26,
    "Beauty_and_Personal_Care": 27,
    "Amazon_Fashion": 28,
    "Magazine_Subscriptions": 29,
    "Software": 30,
    "Health_and_Personal_Care": 31,
    "Appliances": 32,
    "Movies_and_TV": 33,
}
amazon_int_to_str = {v: k for k, v in amazon_category_dict.items()}


amazon_rating_dict = {
    1.0: 1.0,
    5.0: 5.0,
}

dataset_metadata = {
    "LabHC/bias_in_bios": {
        "text_column_name": "hard_text",
        "column1_name": "profession",
        "column2_name": "gender",
        "column2_autointerp_name": "gender",
        "column1_mapping": profession_dict,
        "column2_mapping": gender_dict,
    },
    "canrager/amazon_reviews_mcauley_1and5": {
        "text_column_name": "text",
        "column1_name": "category",
        "column2_name": "rating",
        "column2_autointerp_name": "Amazon Review Sentiment",
        "column1_mapping": amazon_category_dict,
        "column2_mapping": amazon_rating_dict,
    },
}

# These classes are selected as they have at least 4000 samples in the training set when balanced by gender / rating
chosen_classes_per_dataset = {
    "LabHC/bias_in_bios_class_set1": ["0", "1", "2", "6", "9"],
    "LabHC/bias_in_bios_class_set2": ["11", "13", "14", "18", "19"],
    "LabHC/bias_in_bios_class_set3": ["20", "21", "22", "25", "26"],
    "canrager/amazon_reviews_mcauley_1and5": ["1", "2", "3", "5", "6"],
    "canrager/amazon_reviews_mcauley_1and5_sentiment": ["1.0", "5.0"],
    "codeparrot/github-code": ["C", "Python", "HTML", "Java", "PHP"],
    "fancyzhx/ag_news": ["0", "1", "2", "3"],
    "Helsinki-NLP/europarl": ["en", "fr", "de", "es", "nl"],
}



from pydantic import Field
from pydantic.dataclasses import dataclass

from sae_bench.evals.base_eval_output import BaseEvalConfig


@dataclass
class UnlearningEvalConfig(BaseEvalConfig):
    random_seed: int = Field(default=42, title="Random Seed", description="Random seed")

    dataset_names: list[str] = Field(
        default_factory=lambda: [
            "wmdp-bio",
            "high_school_us_history",
            "college_computer_science",
            "high_school_geography",
            "human_aging",
        ],
        title="Dataset Names",
        description="List of dataset names. We want to unlearn wmdp-bio while retaining knowledge in other datasets",
    )

    intervention_method: str = Field(
        default="clamp_feature_activation",
        title="Intervention Method",
        description="Intervention method. We only support 'clamp_feature_activation' for now",
    )

    retain_thresholds: list[float] = Field(
        default_factory=lambda: [0.001, 0.01],
        title="Retain Thresholds",
        description="We ignore features that activate more than this threshold on the retain dataset",
    )
    n_features_list: list[int] = Field(
        default_factory=lambda: [10, 20],
        title="N Features List",
        description="Each N is the number of features we select and clamp to a negative value",
    )
    multipliers: list[int] = Field(
        default_factory=lambda: [25, 50, 100, 200],
        title="Multipliers",
        description="A list of negative values. We iterate over this list, clamping the selected features to each value",
    )

    dataset_size: int = Field(
        default=1024,
        title="Dataset Size",
        description="Dataset size we use when calculating feature sparsity",
    )
    seq_len: int = Field(
        default=1024,
        title="Sequence Length",
        description="Sequence length when calculating feature sparsity",
    )

    n_batch_loss_added: int = Field(
        default=50,
        title="N Batch Loss Added",
        description="Number of batches to use when calculating the loss added by an intervention (currently not supported).",
    )
    target_metric: str = Field(
        default="correct",
        title="Target Metric",
        description="Controls the type of `question_ids` we load. We support 'correct', `correct-iff-question`, and `correct-no-tricks",
    )
    save_metrics: bool = Field(
        default=True,
        title="Save Metrics Flag",
        description="If true, we save the metrics for each set of intervention hyperparameters. This is required to be true currently, as the unlearning score is calculated over all results.",
    )

    model_name: str = Field(
        default="",
        title="Model Name",
        description="Model name. Must be set with a command line argument. We recommend instruct tuned models >= 2B parameters.",
    )
    llm_batch_size: int = Field(
        default=None,
        title="LLM Batch Size",
        description="LLM batch size. This is set by default in the main script, or it can be set with a command line argument.",
    )  # type: ignore
    llm_dtype: str = Field(
        default="",
        title="LLM Data Type",
        description="LLM data type. This is set by default in the main script, or it can be set with a command line argument.",
    )



# type: ignore
# TODO: get_eval_results and get_core_results are broken, these need to be fixed, then turn type checking back on


import json
import os
import re
from collections import defaultdict
from copy import deepcopy
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import seaborn as sns
from matplotlib.colors import Normalize
from matplotlib.lines import Line2D
from scipy import stats

# create a dictionary mapping trainer types to marker shapes

TRAINER_MARKERS = {
    "standard": "o",
    "standard_april_update": "o",
    "jumprelu": "X",
    "topk": "^",
    "batch_topk": "s",
    "p_anneal": "P",
    "matryoshka_batch_topk": "*",
    "gated": "d",
}

TRAINER_COLORS = {
    "standard": "blue",
    "standard_april_update": "blue",
    "jumprelu": "orange",
    "topk": "green",
    "batch_topk": "purple",
    "p_anneal": "red",
    "matryoshka_batch_topk": "brown",
    "gated": "purple",
}


TRAINER_LABELS = {
    "standard": "Standard",
    "standard_april_update": "Standard",
    "jumprelu": "JumpReLU",
    "topk": "TopK",
    "batch_topk": "Batch TopK",
    "p_anneal": "Standard with P-Annealing",
    "matryoshka_batch_topk": "Matryoshka Batch TopK",
    "gated": "Gated",
}


# default text size
plt.rcParams.update({"font.size": 20})


def get_best_results(
    results_dict: dict[str, dict[str, float]], results_path: str, ks: list[int]
) -> dict[str, dict[str, float]]:
    best_results = {}
    for sae, data in results_dict.items():
        best_results[sae] = 0
        for k in ks:
            custom_metric, _ = get_custom_metric_key_and_name(results_path, k)
            if custom_metric in data:
                best_results[sae] = max(best_results[sae], data[custom_metric])
            else:
                print(f"Custom metric {custom_metric} not found for {sae}")

    for sae in best_results.keys():
        results_dict[sae]["best_custom_metric"] = best_results[sae]

    return results_dict


def get_single_figure(
    selected_saes: list[tuple[str, str]],
    results_path: str,
    core_results_path: str,
    image_base_name: str,
    k: int | None = None,
    trainer_markers: dict[str, str] | None = None,
    title: str | None = None,
    title_prefix: str = "",
    plot_type: bool = True,
    baseline_sae: tuple[str, str] | None = None,
    baseline_label: str | None = None,
):
    eval_results = get_eval_results(selected_saes, results_path)
    core_results = get_core_results(selected_saes, core_results_path)

    for sae in eval_results:
        eval_results[sae].update(core_results[sae])

    custom_metric, custom_metric_name = get_custom_metric_key_and_name(results_path, k)

    if baseline_sae:
        baseline_results = get_eval_results([baseline_sae], results_path)
        baseline_id = f"{baseline_sae[0]}_{baseline_sae[1]}"
        baseline_results[baseline_id].update(
            get_core_results([baseline_sae], core_results_path)[baseline_id]
        )
        baseline_value = baseline_results[baseline_id][custom_metric]
        assert baseline_label, "Please provide a label for the baseline"
    else:
        baseline_value = None
        assert baseline_label is None, "Please do not provide a label for the baseline"

    if not title:
        title = f"{title_prefix}L0 vs {custom_metric_name}"

    if plot_type:
        fig = plot_2var_graph(
            eval_results,
            custom_metric,
            y_label=custom_metric_name,
            title=title,
            output_filename=f"{image_base_name}_2var_sae_type.png",
            trainer_markers=trainer_markers,
            return_fig=True,
            baseline_value=baseline_value,
            baseline_label=baseline_label,
        )
    else:
        fig = plot_2var_graph_dict_size(
            eval_results,
            custom_metric,
            y_label=custom_metric_name,
            title=title,
            output_filename=f"{image_base_name}_2var_dict_size.png",
            return_fig=True,
            baseline_value=baseline_value,
            baseline_label=baseline_label,
        )

    return fig


def plot_results(
    eval_filenames: list[str],
    core_filenames: list[str],
    eval_type: str,
    image_base_name: str,
    k: int | None = None,
    trainer_markers: dict[str, str] | None = None,
    trainer_colors: dict[str, str] | None = None,
    title_prefix: str = "",
    return_fig: bool = False,
    baseline_sae_path: str | None = None,
    baseline_label: str | None = None,
):
    eval_results = get_eval_results(eval_filenames)
    core_results = get_core_results(core_filenames)

    for sae in eval_results:
        eval_results[sae].update(core_results[sae])

    custom_metric, custom_metric_name = get_custom_metric_key_and_name(eval_type, k)

    if baseline_sae_path:
        baseline_results = get_eval_results([baseline_sae_path])

        baseline_filename = os.path.basename(baseline_sae_path)
        baseline_results_key = baseline_filename.replace("_eval_results.json", "")

        core_baseline_filename = baseline_sae_path.replace(eval_type, "core")

        baseline_results[baseline_results_key].update(
            get_core_results([core_baseline_filename])[baseline_results_key]
        )

        baseline_value = baseline_results[baseline_results_key][custom_metric]
        assert baseline_label, "Please provide a label for the baseline"
    else:
        baseline_value = None
        assert baseline_label is None, "Please do not provide a label for the baseline"

    # title_3var = f"{title_prefix}L0 vs Loss Recovered vs {custom_metric_name}"
    # plot_3var_graph(
    #     eval_results,
    #     title_3var,
    #     custom_metric,
    #     colorbar_label="Custom Metric",
    #     output_filename=f"{image_base_name}_3var.png",
    #     trainer_markers=trainer_markers,
    # )

    title_2var = f"{title_prefix}L0 vs {custom_metric_name}"
    fig = plot_2var_graph(
        eval_results,
        custom_metric,
        y_label=custom_metric_name,
        title=title_2var,
        output_filename=f"{image_base_name}_2var_sae_type.png",
        trainer_markers=trainer_markers,
        trainer_colors=trainer_colors,
        baseline_value=baseline_value,
        baseline_label=baseline_label,
    )

    plot_2var_graph_dict_size(
        eval_results,
        custom_metric,
        y_label=custom_metric_name,
        title=title_2var,
        output_filename=f"{image_base_name}_2var_dict_size.png",
        baseline_value=baseline_value,
        baseline_label=baseline_label,
        trainer_markers=trainer_markers,
    )

    if return_fig:
        return fig


def plot_best_of_ks_results(
    selected_saes: list[tuple[str, str]],
    results_path: str,
    core_results_path: str,
    image_base_name: str,
    ks: list[int],
    trainer_markers: dict[str, str] | None = None,
    title_prefix: str = "",
    baseline_sae: tuple[str, str] | None = None,
    baseline_label: str | None = None,
):
    dummy_k = 0

    eval_results = get_eval_results(selected_saes, results_path)
    core_results = get_core_results(selected_saes, core_results_path)

    for sae in eval_results:
        eval_results[sae].update(core_results[sae])

    custom_metric, custom_metric_name = get_custom_metric_key_and_name(
        results_path, dummy_k
    )

    custom_metric = "best_custom_metric"
    custom_metric_name = custom_metric_name.replace(f"Top {dummy_k}", f"Best of {ks}")
    eval_results = get_best_results(eval_results, results_path, ks)

    if baseline_sae:
        baseline_results = get_eval_results([baseline_sae], results_path)
        baseline_id = f"{baseline_sae[0]}_{baseline_sae[1]}"
        baseline_results[baseline_id].update(
            get_core_results([baseline_sae], core_results_path)[baseline_id]
        )

        baseline_results = get_best_results(baseline_results, results_path, ks)
        baseline_value = baseline_results[baseline_id]["best_custom_metric"]
        assert baseline_label, "Please provide a label for the baseline"
    else:
        baseline_value = None
        assert baseline_label is None, "Please do not provide a label for the baseline"

    title_3var = f"{title_prefix}L0 vs Loss Recovered vs {custom_metric_name}"
    title_2var = f"{title_prefix}L0 vs {custom_metric_name}"

    plot_3var_graph(
        eval_results,
        title_3var,
        custom_metric,
        colorbar_label="Custom Metric",
        output_filename=f"{image_base_name}_3var.png",
        trainer_markers=trainer_markers,
    )
    plot_2var_graph(
        eval_results,
        custom_metric,
        y_label=custom_metric_name,
        title=title_2var,
        output_filename=f"{image_base_name}_2var_sae_type.png",
        trainer_markers=trainer_markers,
        baseline_value=baseline_value,
        baseline_label=baseline_label,
    )

    plot_2var_graph_dict_size(
        eval_results,
        custom_metric,
        y_label=custom_metric_name,
        title=title_2var,
        output_filename=f"{image_base_name}_2var_dict_size.png",
        baseline_value=baseline_value,
        baseline_label=baseline_label,
    )


def get_custom_metric_key_and_name(
    eval_path: str, k: int | None = None
) -> tuple[str, str]:
    if "tpp" in eval_path:
        custom_metric = f"tpp_threshold_{k}_total_metric"
        custom_metric_name = f"TPP Top {k} Metric"
    elif "scr" in eval_path:
        custom_metric = f"scr_metric_threshold_{k}"
        custom_metric_name = f"SCR Top {k} Metric"
    elif "sparse_probing" in eval_path:
        custom_metric = f"sae_top_{k}_test_accuracy"
        custom_metric_name = f"Sparse Probing Top {k} Test Accuracy"
    elif "absorption" in eval_path:
        custom_metric = "mean_absorption_fraction_score"
        custom_metric_name = "Mean Absorption Score"
    elif "autointerp" in eval_path:
        custom_metric = "autointerp_score"
        custom_metric_name = "Autointerp Score"
    elif "unlearning" in eval_path:
        custom_metric = "unlearning_score"
        custom_metric_name = "Unlearning Score"
    elif "core" in eval_path:
        custom_metric = "ce_loss_score"
        custom_metric_name = "Loss Recovered"
    else:
        raise ValueError("Please add the correct key for the custom metric")

    return custom_metric, custom_metric_name


def get_sae_bench_train_tokens(filename: str) -> int:
    """This is for SAE Bench internal use. The SAE cfg does not contain the number of training tokens, so we need to hardcode it."""

    if "sae_bench" not in filename:
        raise ValueError("This function is only for SAE Bench releases")

    batch_size = 2048

    if "step" not in filename:
        steps = 244140
        return steps * batch_size
    else:
        match = re.search(r"step_(\d+)", filename)
        if match:
            step = int(match.group(1))
            return step * batch_size
        else:
            raise ValueError("No step match found")


def get_d_sae_string(d_sae: int) -> str:
    rounded_d_sae = round(d_sae / 1000) * 1000

    # TODO: Temp SAE Bench fix
    if rounded_d_sae == 66000:
        rounded_d_sae = 65000

    if rounded_d_sae >= 1_000_000 and rounded_d_sae <= 1_060_000:
        return "1M"
    else:
        return f"{rounded_d_sae // 1000}k"


def get_eval_results(eval_filenames: list[str]) -> dict[str, dict]:
    """eval_filenames is assumed to be a list of filenames of this format:
    {sae_release}_{sae_id}_eval_results.json"""
    eval_results = {}
    for filepath in eval_filenames:
        if not os.path.exists(filepath):
            print(f"File not found: {filepath}")
            continue

        with open(filepath) as f:
            single_sae_results = json.load(f)

        filename = os.path.basename(filepath)
        results_key = filename.replace("_eval_results.json", "")

        if "tpp" in filepath:
            eval_results[results_key] = single_sae_results["eval_result_metrics"][
                "tpp_metrics"
            ]
        elif "scr" in filepath:
            eval_results[results_key] = single_sae_results["eval_result_metrics"][
                "scr_metrics"
            ]
        elif "absorption" in filepath:
            eval_results[results_key] = single_sae_results["eval_result_metrics"][
                "mean"
            ]
        elif "autointerp" in filepath:
            eval_results[results_key] = single_sae_results["eval_result_metrics"][
                "autointerp"
            ]
        elif "sparse_probing" in filepath:
            eval_results[results_key] = single_sae_results["eval_result_metrics"]["sae"]
        elif "unlearning" in filepath:
            eval_results[results_key] = single_sae_results["eval_result_metrics"][
                "unlearning"
            ]
        elif "core" in filepath:
            # core has nested evaluation metrics, so we flatten them out here
            core_results = single_sae_results["eval_result_metrics"]
            eval_results[results_key] = {}
            for parent_key, child_dict in core_results.items():
                for metric_key, value in child_dict.items():
                    eval_results[results_key][metric_key] = value
        else:
            raise ValueError("Please add the correct key for the eval results")

        eval_results[results_key]["eval_config"] = single_sae_results["eval_config"]

        sae_config = single_sae_results["sae_cfg_dict"]

        eval_results[results_key]["sae_class"] = sae_config["architecture"]

        eval_results[results_key]["d_sae"] = get_d_sae_string(sae_config["d_sae"])

        if "training_tokens" in sae_config:
            eval_results[results_key]["train_tokens"] = sae_config["training_tokens"]
        else:
            eval_results[results_key]["train_tokens"] = 1e-6

    return eval_results


def get_core_results(core_filenames: list[str]) -> dict:
    core_results = {}
    for filepath in core_filenames:
        if not os.path.exists(filepath):
            print(f"File not found: {filepath}")
            continue

        with open(filepath) as f:
            single_sae_results = json.load(f)

        l0 = single_sae_results["eval_result_metrics"]["sparsity"]["l0"]
        ce_score = single_sae_results["eval_result_metrics"][
            "model_performance_preservation"
        ]["ce_loss_score"]

        filename = os.path.basename(filepath)
        results_key = filename.replace("_eval_results.json", "")
        core_results[results_key] = {"l0": l0, "ce_loss_score": ce_score}
    return core_results


def find_eval_results_files(folders: list[str]) -> list[str]:
    """
    Recursively explores the given list of folder names and finds all file paths
    containing 'eval_results.json'. Returns a list of the full file paths.

    Args:
        folders (list[str]): A list of folder names to explore.

    Returns:
        list[str]: A list of full file paths containing 'eval_results.json'.
    """
    result_files = []

    for folder in folders:
        for root, dirs, files in os.walk(folder):
            for file in files:
                if "eval_results.json" in file:
                    result_files.append(os.path.join(root, file))

    return result_files


def update_trainer_markers_and_colors(
    results: dict[str, dict[str, Any]],
    trainer_markers: dict[str, str],
    trainer_colors: dict[str, str],
) -> tuple[dict[str, str], dict[str, str]]:
    trainer_markers = deepcopy(trainer_markers)
    trainer_colors = deepcopy(trainer_colors)

    available_markers = list(set(trainer_markers.values()))
    available_colors = list(set(trainer_colors.values()))

    existing_trainers = set(trainer_markers.keys())
    all_trainers = {v["sae_class"] for v in results.values()}
    new_trainers = all_trainers - existing_trainers

    for trainer in all_trainers:
        if trainer in trainer_markers:
            if trainer_markers[trainer] in available_markers:
                available_markers.remove(trainer_markers[trainer])
            if trainer_colors[trainer] in available_colors:
                available_colors.remove(trainer_colors[trainer])

    for trainer in new_trainers:
        trainer_markers[trainer] = available_markers.pop(0)
        trainer_colors[trainer] = available_colors.pop(0)

    return trainer_markers, trainer_colors


def plot_3var_graph(
    results: dict[str, dict[str, float]],
    title: str,
    custom_metric: str,
    xlims: tuple[float, float] | None = None,
    ylims: tuple[float, float] | None = None,
    colorbar_label: str = "Average Diff",
    output_filename: str | None = None,
    legend_location: str = "lower right",
    x_axis_key: str = "l0",
    y_axis_key: str = "ce_loss_score",
    trainer_markers: dict[str, str] | None = None,
):
    if not trainer_markers:
        trainer_markers = TRAINER_MARKERS

    trainer_markers, _ = update_trainer_markers_and_colors(
        results, trainer_markers, TRAINER_COLORS
    )

    # Extract data from results
    l0_values = [data[x_axis_key] for data in results.values()]
    frac_recovered_values = [data[y_axis_key] for data in results.values()]
    custom_metric_values = [data[custom_metric] for data in results.values()]

    # Create the scatter plot
    fig, ax = plt.subplots(figsize=(10, 6))

    # Create a normalize object for color scaling
    norm = Normalize(vmin=min(custom_metric_values), vmax=max(custom_metric_values))

    handles, labels = [], []

    for trainer, marker in trainer_markers.items():
        # Filter data for this trainer
        trainer_data = {k: v for k, v in results.items() if v["sae_class"] == trainer}

        if not trainer_data:
            continue  # Skip this trainer if no data points

        l0_values = [data[x_axis_key] for data in trainer_data.values()]
        frac_recovered_values = [data[y_axis_key] for data in trainer_data.values()]
        custom_metric_values = [data[custom_metric] for data in trainer_data.values()]

        # Plot data points
        scatter = ax.scatter(
            l0_values,
            frac_recovered_values,
            c=custom_metric_values,
            cmap="viridis",
            marker=marker,
            s=100,
            label=trainer,
            norm=norm,
            edgecolor="black",
        )

        # custom legend stuff
        _handle, _ = scatter.legend_elements(prop="sizes")
        _handle[0].set_markeredgecolor("black")
        _handle[0].set_markerfacecolor("white")
        _handle[0].set_markersize(10)
        if marker == "d":
            _handle[0].set_markersize(13)
        handles += _handle

        if trainer in TRAINER_LABELS:
            trainer_label = TRAINER_LABELS[trainer]
        else:
            trainer_label = trainer.capitalize()

        labels.append(trainer_label)

    # Add colorbar
    fig.colorbar(scatter, ax=ax, label=colorbar_label)  # type: ignore

    # Set labels and title
    ax.set_xlabel("L0 (Sparsity)")
    ax.set_ylabel("Loss Recovered (Fidelity)")
    ax.set_title(title)

    ax.legend(handles, labels, loc=legend_location)

    # Set axis limits
    if xlims:
        ax.set_xlim(*xlims)
    if ylims:
        ax.set_ylim(*ylims)

    plt.tight_layout()

    # Save and show the plot
    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")
    plt.show()


def plot_interactive_3var_graph(
    results: dict[str, dict[str, float]],
    custom_color_metric: str,
    xlims: tuple[float, float] | None = None,
    y_lims: tuple[float, float] | None = None,
    output_filename: str | None = None,
    x_axis_key: str = "l0",
    y_axis_key: str = "ce_loss_score",
    title: str = "",
):
    # Extract data from results
    ae_paths = list(results.keys())
    l0_values = [data[x_axis_key] for data in results.values()]
    frac_recovered_values = [data[y_axis_key] for data in results.values()]

    custom_metric_value = [data[custom_color_metric] for data in results.values()]

    dict_size = [data["d_sae"] for data in results.values()]

    # Create the scatter plot
    fig = go.Figure()

    # Add trace
    fig.add_trace(
        go.Scatter(
            x=l0_values,
            y=frac_recovered_values,
            mode="markers",
            marker=dict(
                size=10,
                color=custom_metric_value,  # Color points based on frac_recovered
                colorscale="Viridis",  # You can change this colorscale
                showscale=True,
            ),
            text=[
                f"AE Path: {ae}<br>L0: {l0:.4f}<br>Frac Recovered: {fr:.4f}<br>Custom Metric: {ad:.4f}<br>Dict Size: {d}"
                for ae, l0, fr, ad, d in zip(
                    ae_paths,
                    l0_values,
                    frac_recovered_values,
                    custom_metric_value,
                    dict_size,
                )
            ],
            hoverinfo="text",
        )
    )

    # Update layout
    fig.update_layout(
        title=title,
        xaxis_title="L0 (Sparsity)",
        yaxis_title="Loss Recovered (Fidelity)",
        hovermode="closest",
    )

    # Set axis limits
    if xlims:
        fig.update_xaxes(range=xlims)
    if y_lims:
        fig.update_yaxes(range=y_lims)

    # Save and show the plot
    if output_filename:
        fig.write_html(output_filename)

    fig.show()


def plot_2var_graph(
    results: dict[str, dict[str, float]],
    custom_metric: str,
    title: str = "L0 vs Custom Metric",
    y_label: str = "Custom Metric",
    xlims: tuple[float, float] | None = None,
    ylims: tuple[float, float] | None = None,
    output_filename: str | None = None,
    baseline_value: float | None = None,
    baseline_label: str | None = None,
    x_axis_key: str = "l0",
    trainer_markers: dict[str, str] | None = None,
    trainer_colors: dict[str, str] | None = None,
    return_fig: bool = False,
    connect_points: bool = False,  # New parameter to control line connections
):
    if not trainer_markers:
        trainer_markers = TRAINER_MARKERS

    if not trainer_colors:
        trainer_colors = TRAINER_COLORS

    trainer_markers, trainer_colors = update_trainer_markers_and_colors(
        results, trainer_markers, trainer_colors
    )

    # Create the scatter plot with extra width for legend
    fig, ax = plt.subplots(figsize=(12, 6))

    handles, labels = [], []

    for trainer, marker in trainer_markers.items():
        # Filter data for this trainer
        trainer_data = {k: v for k, v in results.items() if v["sae_class"] == trainer}

        if not trainer_data:
            continue  # Skip this trainer if no data points

        l0_values = [data[x_axis_key] for data in trainer_data.values()]
        custom_metric_values = [data[custom_metric] for data in trainer_data.values()]

        # Sort points by l0 values for proper line connection
        if connect_points and len(l0_values) > 1:
            points = sorted(zip(l0_values, custom_metric_values))
            l0_values = [p[0] for p in points]
            custom_metric_values = [p[1] for p in points]

            # Add connecting line
            ax.plot(
                l0_values,
                custom_metric_values,
                color=trainer_colors[trainer],
                linestyle="-",
                alpha=0.5,
                zorder=1,  # Ensure lines are plotted behind points
            )

        # Plot data points
        ax.scatter(
            l0_values,
            custom_metric_values,
            marker=marker,
            s=100,
            label=trainer,
            color=trainer_colors[trainer],
            edgecolor="black",
            zorder=2,  # Ensure points are plotted on top of lines
        )

        # Create custom legend handle with both marker and color
        legend_handle = plt.scatter(
            [],
            [],
            marker=marker,
            s=100,
            color=trainer_colors[trainer],
            edgecolor="black",
        )
        handles.append(legend_handle)

        if trainer in TRAINER_LABELS:
            trainer_label = TRAINER_LABELS[trainer]
        else:
            trainer_label = trainer.capitalize()
        labels.append(trainer_label)

    # Set labels and title
    ax.set_xlabel("L0 (Sparsity)")
    ax.set_ylabel(y_label)
    ax.set_title(title)

    # x log
    ax.set_xscale("log")

    if baseline_value is not None:
        ax.axhline(baseline_value, color="red", linestyle="--", label=baseline_label)
        labels.append(baseline_label)
        handles.append(
            Line2D([0], [0], color="red", linestyle="--", label=baseline_label)
        )

    # Place legend outside the plot on the right
    ax.legend(handles, labels, loc="center left", bbox_to_anchor=(1, 0.5))

    # Set axis limits
    if xlims:
        ax.set_xlim(*xlims)
    if ylims:
        ax.set_ylim(*ylims)

    plt.tight_layout()

    # Save and show the plot
    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")

    if return_fig:
        return fig

    plt.show()


def plot_2var_graph_dict_size(
    results: dict[str, dict[str, float]],
    custom_metric: str,
    title: str = "L0 vs Custom Metric",
    y_label: str = "Custom Metric",
    xlims: tuple[float, float] | None = None,
    ylims: tuple[float, float] | None = None,
    output_filename: str | None = None,
    legend_location: str = "lower right",
    baseline_value: float | None = None,
    baseline_label: str | None = None,
    x_axis_key: str = "l0",
    return_fig: bool = False,
    trainer_markers: dict[str, str] | None = None,
):
    if not trainer_markers:
        trainer_markers = TRAINER_MARKERS

    # Extract data
    l0_values = [data[x_axis_key] for data in results.values()]
    custom_metric_values = [data[custom_metric] for data in results.values()]

    # Create the scatter plot
    fig, ax = plt.subplots(figsize=(10, 6))

    # Define possible dict sizes and their markers
    possible_sizes = ["4k", "16k", "65k", "131k", "1M"]

    # Create color map with more exaggerated differences for 3 colors
    colors = [plt.cm.Reds(x) for x in [0.1, 0.5, 0.9]]  # type: ignore # Light, medium, dark red

    # Get unique dict sizes present in the data while preserving order from possible_sizes
    unique_sizes = [
        size
        for size in possible_sizes
        if size in set(v["d_sae"] for v in results.values())
    ]

    assert len(unique_sizes) <= len(colors), (
        "Too many unique dictionary sizes for color map"
    )

    size_to_color = {size: colors[i] for i, size in enumerate(unique_sizes)}

    # Iterate over each unique dictionary size
    handles, labels = [], []

    for dict_size in unique_sizes:
        # Filter data points for the current dictionary size
        size_data = {k: v for k, v in results.items() if v["d_sae"] == dict_size}

        # Get values for l0 and custom metric for this dictionary size
        l0_values = [data[x_axis_key] for data in size_data.values()]
        custom_metric_values = [data[custom_metric] for data in size_data.values()]
        sae_classes = [data["sae_class"] for data in size_data.values()]

        # Plot data points with the assigned marker and color
        for l0, metric, sae_class in zip(l0_values, custom_metric_values, sae_classes):
            marker = trainer_markers[sae_class]  # type: ignore
            ax.scatter(
                l0,
                metric,
                marker=marker,
                s=100,
                color=size_to_color[dict_size],
                edgecolor="black",
            )

        # Collect legend handles and labels
        _handle = plt.scatter(
            [], [], marker="o", s=100, color=size_to_color[dict_size], edgecolor="black"
        )
        handles.append(_handle)
        labels.append(f"SAE Width: {dict_size}")

    # Set labels and title
    ax.set_xlabel("L0 (Sparsity)")
    ax.set_ylabel(y_label)
    ax.set_title(title)

    if baseline_value:
        ax.axhline(baseline_value, color="red", linestyle="--", label=baseline_label)
        labels.append(baseline_label)
        handles.append(
            Line2D([0], [0], color="red", linestyle="--", label=baseline_label)
        )

    ax.legend(handles, labels, loc=legend_location)

    # Set axis limits
    if xlims:
        ax.set_xlim(*xlims)
    if ylims:
        ax.set_ylim(*ylims)

    # log scale
    ax.set_xscale("log")

    plt.tight_layout()

    # Save and show the plot
    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")

    if return_fig:
        return fig
    plt.show()


def plot_steps_vs_average_diff(
    results_dict: dict,
    steps_key: str = "train_tokens",
    avg_diff_key: str = "average_diff",
    title: str | None = None,
    y_label: str | None = None,
    output_filename: str | None = None,
):
    # Initialize a defaultdict to store data for each trainer
    trainer_data = defaultdict(lambda: {"train_tokens": [], "metric_scores": []})

    all_steps = set()

    # Extract data from the dictionary
    for key, value in results_dict.items():
        # Extract trainer number from the key
        trainer = key.split("/")[-1].split("_")[
            1
        ]  # Assuming format like "trainer_1_..."
        layer = key.split("/")[-2].split("_")[-2]

        if "topk_ctx128" in key:
            trainer_type = "TopK SAE"
        elif "standard_ctx128" in key:
            trainer_type = "Standard SAE"
        else:
            raise ValueError(f"Unknown trainer type in key: {key}")

        step = int(value[steps_key])
        avg_diff = value[avg_diff_key]

        trainer_key = f"{trainer_type} Layer {layer} Trainer {trainer}"

        trainer_data[trainer_key]["train_tokens"].append(step)
        trainer_data[trainer_key]["metric_scores"].append(avg_diff)
        all_steps.add(step)

    # Calculate average across all trainers
    average_trainer_data = {"train_tokens": [], "metric_scores": []}
    for step in sorted(all_steps):
        step_diffs = []
        for data in trainer_data.values():
            if step in data["train_tokens"]:
                idx = data["train_tokens"].index(step)
                step_diffs.append(data["metric_scores"][idx])
        if step_diffs:
            average_trainer_data["train_tokens"].append(step)
            average_trainer_data["metric_scores"].append(np.mean(step_diffs))

    # Add average_trainer_data to trainer_data
    trainer_data["Average"] = average_trainer_data

    # Create the plot
    plt.figure(figsize=(12, 6))

    # Plot data for each trainer
    for trainer_key, data in trainer_data.items():
        steps = data["train_tokens"]
        metric_scores = data["metric_scores"]

        # Sort the data by steps to ensure proper ordering
        sorted_data = sorted(zip(steps, metric_scores))
        steps, metric_scores = zip(*sorted_data)

        # Find the maximum step value for this trainer
        max_step = max(steps)

        # Convert steps to percentages of max_step
        step_percentages = [step / max_step * 100 for step in steps]

        # Plot the line for this trainer
        if trainer_key == "Average":
            plt.plot(
                step_percentages,
                metric_scores,
                marker="o",
                label=trainer_key,
                linewidth=3,
                color="red",
                zorder=10,
            )  # Emphasized average line
        else:
            plt.plot(
                step_percentages,
                metric_scores,
                marker="o",
                label=trainer_key,
                alpha=0.3,
                linewidth=1,
            )  # More transparent individual lines

    # log scale
    # plt.xscale("log")

    # if not title:
    #     title = f'{steps_key.capitalize()} vs {avg_diff_key.replace("_", " ").capitalize()}'

    if not y_label:
        y_label = avg_diff_key.replace("_", " ").capitalize()

    plt.xlabel("Training Progess (%)")
    plt.ylabel(y_label)
    plt.title(title)
    plt.grid(True, alpha=0.3)  # More transparent grid

    if len(trainer_data) < 50 and False:
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0.0)

    # Adjust layout to prevent clipping of tick-labels
    plt.tight_layout()

    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")

    # Show the plot
    plt.show()


def plot_correlation_heatmap(
    plotting_results: dict[str, dict[str, float]],
    metric_names: list[str],
    ae_names: list[str] | None = None,
    title: str = "Metric Correlation Heatmap",
    output_filename: str = None,
    figsize: tuple = (12, 10),
    cmap: str = "coolwarm",
    annot: bool = True,
):
    # If ae_names is not provided, use all ae_names from plotting_results
    if ae_names is None:
        ae_names = list(plotting_results.keys())

    # If metric_names is not provided, use all metric names from the first ae_name
    # if metric_names is None:
    #     metric_names = list(plotting_results[ae_names[0]].keys())

    # Create a DataFrame from the plotting_results
    data = []
    for ae in ae_names:
        row = [plotting_results[ae].get(metric, np.nan) for metric in metric_names]
        data.append(row)

    df = pd.DataFrame(data, index=ae_names, columns=metric_names)

    # Calculate the correlation matrix
    corr_matrix = df.corr()

    # Create the heatmap
    plt.figure(figsize=figsize)
    sns.heatmap(corr_matrix, annot=annot, cmap=cmap, vmin=-1, vmax=1, center=0)

    plt.title(title)
    plt.tight_layout()

    # Save the plot if output_filename is provided
    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")

    plt.show()


def plot_correlation_scatter(
    plotting_results: dict[str, dict[str, float]],
    metric_x: str,
    metric_y: str,
    x_label: str | None = None,
    y_label: str | None = None,
    ae_names: list[str] | None = None,
    title: str = "Metric Comparison Scatter Plot",
    output_filename: str | None = None,
    figsize: tuple = (10, 8),
):
    # If ae_names is not provided, use all ae_names from plotting_results
    if ae_names is None:
        ae_names = list(plotting_results.keys())

    # Extract x and y values for the specified metrics
    x_values = [plotting_results[ae].get(metric_x, float("nan")) for ae in ae_names]
    y_values = [plotting_results[ae].get(metric_y, float("nan")) for ae in ae_names]

    # Remove any NaN values
    valid_data = [
        (x, y, ae)
        for x, y, ae in zip(x_values, y_values, ae_names)
        if not (np.isnan(x) or np.isnan(y))
    ]
    if not valid_data:
        print("No valid data points after removing NaN values.")
        return

    x_values, y_values, valid_ae_names = zip(*valid_data)

    # Convert to numpy arrays
    x_values = np.array(x_values)
    y_values = np.array(y_values)

    # Calculate correlation coefficients
    r, p_value = stats.pearsonr(x_values, y_values)
    r_squared = r**2  # type: ignore

    # Create the scatter plot
    plt.figure(figsize=figsize)
    sns.scatterplot(x=x_values, y=y_values, label="SAE", color="blue")

    if x_label is None:
        x_label = metric_x
    if y_label is None:
        y_label = metric_y

    # Add labels and title
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)

    # Add a trend line
    sns.regplot(
        x=x_values, y=y_values, scatter=False, color="red", label=f"r = {r:.4f}"
    )

    plt.legend()

    plt.tight_layout()

    # Save the plot if output_filename is provided
    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")

    plt.show()

    # Print correlation coefficients
    print(f"Pearson correlation coefficient (r): {r:.4f}")
    print(f"Coefficient of determination (r²): {r_squared:.4f}")
    print(f"P-value: {p_value:.4f}")


def plot_training_steps(
    results_dict: dict,
    metric_key: str,
    steps_key: str = "train_tokens",
    title: str | None = None,
    y_label: str | None = None,
    output_filename: str | None = None,
    break_fraction: float = 0.15,  # Parameter to control break position
):
    # Initialize a defaultdict to store data for each trainer
    trainer_data = defaultdict(lambda: {"train_tokens": [], "metric_scores": []})
    all_steps = set()
    all_trainers = set()

    # Extract data from the dictionary
    for key, value in results_dict.items():
        trainer = key.split("_trainer_")[-1].split("_")[0]
        trainer_class = value["sae_class"]
        step = int(value[steps_key])
        metric_scores = value[metric_key]
        trainer_key = f"{trainer_class} Trainer {trainer}"

        trainer_data[trainer_key]["train_tokens"].append(step)
        trainer_data[trainer_key]["metric_scores"].append(metric_scores)
        trainer_data[trainer_key]["sae_class"] = trainer_class
        all_steps.add(step)
        all_trainers.add(trainer_class)

    # Calculate average across all trainers
    average_trainer_data = {"train_tokens": [], "metric_scores": []}
    for step in sorted(all_steps):
        step_diffs = [
            data["metric_scores"][data["train_tokens"].index(step)]
            for data in trainer_data.values()
            if step in data["train_tokens"]
        ]
        if step_diffs:
            average_trainer_data["train_tokens"].append(step)
            average_trainer_data["metric_scores"].append(np.mean(step_diffs))
    trainer_data["Average"] = average_trainer_data

    # Create the plot with broken axis
    fig, (ax1, ax2) = plt.subplots(
        1,
        2,
        sharey=True,
        figsize=(15, 6),
        gridspec_kw={"width_ratios": [break_fraction, 1 - break_fraction]},
    )
    fig.subplots_adjust(wspace=0.01)  # Adjust space between axes

    # Calculate break point based on data
    steps_break_point = min([s for s in all_steps if s > 0]) / 2
    break_point = steps_break_point  # / max(all_steps) * 100  # Convert to percentage

    for trainer_key, data in trainer_data.items():
        steps = data["train_tokens"]
        metric_scores = data["metric_scores"]

        if trainer_key == "Average":
            color, trainer_class = "black", "Average"
        elif (
            data["sae_class"] == "standard"
            or data["sae_class"] == "Vanilla"
            or data["sae_class"] == "standard_april_update"
        ):
            color, trainer_class = "red", data["sae_class"]
        elif data["sae_class"] == "topk":
            color, trainer_class = "blue", data["sae_class"]
        else:
            raise ValueError(f"Trainer type not recognized for {trainer_key}")

        sorted_data = sorted(zip(steps, metric_scores))
        steps, metric_scores = zip(*sorted_data)

        ax1.plot(
            steps,
            metric_scores,
            marker="o",
            label=trainer_class,
            linewidth=4 if trainer_key == "Average" else 2,
            color=color,
            alpha=1 if trainer_key == "Average" else 0.3,
            zorder=10 if trainer_key == "Average" else 1,
        )
        ax2.plot(
            steps,
            metric_scores,
            marker="o",
            label=trainer_class,
            linewidth=4 if trainer_key == "Average" else 2,
            color=color,
            alpha=1 if trainer_key == "Average" else 0.3,
            zorder=10 if trainer_key == "Average" else 1,
        )

    # Set up the broken axis
    ax1.set_xlim(-break_point / 4, break_point)
    # ax2.set_xlim(break_point, 100)
    ax2.set_xscale("log")

    # Hide the spines between ax1 and ax2
    ax1.spines["right"].set_visible(False)
    ax2.spines["left"].set_visible(False)
    ax1.yaxis.tick_left()
    ax2.yaxis.tick_right()
    ax2.yaxis.set_label_position("right")

    # Add break lines
    d = 0.015  # Size of diagonal lines
    kwargs = dict(transform=ax1.transAxes, color="k", clip_on=False, lw=4)

    ax1.plot((1, 1), (-d, +d), **kwargs)  # top-right vertical
    ax1.plot((1, 1), (1 - d, 1 + d), **kwargs)  # bottom-right vertical
    kwargs.update(transform=ax2.transAxes)
    ax2.plot((0, 0), (-d, +d), **kwargs)  # top-left vertical
    ax2.plot((0, 0), (1 - d, 1 + d), **kwargs)  # bottom-left vertical

    # Set labels and title
    if not y_label:
        y_label = metric_key.replace("_", " ").capitalize()
    ax1.set_ylabel(y_label)
    fig.text(0.5, 0.01, "Training Steps", ha="center", va="center")
    if title is not None:
        fig.suptitle(title)

    # Adjust x-axis ticks
    # ax1.set_xticks([0])
    # ax1.set_xticklabels(['0%'])
    # ax2.set_xticks([0.1, 1, 10, 100])
    # ax2.set_xticklabels([f'0.1%', '1%', '10%', '100%'])

    # Add grid
    ax1.grid(True, alpha=0.3)
    ax2.grid(True, alpha=0.3)

    # Add custom legend
    legend_elements = []
    legend_elements.append(Line2D([0], [0], color="black", lw=3, label="Average"))
    if "standard" in all_trainers or "Vanilla" in all_trainers:
        legend_elements.append(Line2D([0], [0], color="red", lw=3, label="Standard"))
    if "topk" in all_trainers:
        legend_elements.append(Line2D([0], [0], color="blue", lw=3, label="TopK"))
    ax2.legend(handles=legend_elements, loc="lower right")

    plt.tight_layout()

    if output_filename:
        plt.savefig(output_filename, bbox_inches="tight")

    plt.show()


def get_sae_class_archived(sae_cfg: dict, sae_release) -> str:
    """For results pre Jan 2025"""
    if "sae_bench" in sae_release and "panneal" in sae_release:
        return "p_anneal"

    if sae_cfg["activation_fn_str"] == "topk":
        return "topk"

    return sae_cfg["architecture"]


def get_sae_bench_train_tokens_archived(sae_release: str, sae_id: str) -> int:
    """For results pre Jan 2025.
    This is for SAE Bench internal use. The SAE cfg does not contain the number of training tokens, so we need to hardcode it."""

    if "sae_bench" not in sae_release:
        raise ValueError("This function is only for SAE Bench releases")

    if "pythia" in sae_release:
        batch_size = 4096
    else:
        batch_size = 2048

    if "step" not in sae_id:
        if "pythia" in sae_release:
            steps = 48828
        elif "2pow14" in sae_release:
            steps = 146484
        elif "2pow12" or "2pow16" in sae_release:
            steps = 97656
        else:
            raise ValueError(f"sae release {sae_release} not recognized")

        return steps * batch_size
    else:
        match = re.search(r"step_(\d+)", sae_id)
        if match:
            step = int(match.group(1))
            return step * batch_size
        else:
            raise ValueError("No step match found")



from dataclasses import dataclass, field


@dataclass
class MDLEvalConfig:
    k_values: list[int] | None = field(default_factory=lambda: [16, 24, 32])
    num_bins_values: list[int] = field(default_factory=lambda: [4, 6, 8, 12, 16, 32])

    random_seed: int = 42
    dataset_name: str = "HuggingFaceFW/fineweb"

    context_length: int = 128

    sae_batch_size: int = 64

    model_name: str = "pythia-70m-deduped"
    llm_dtype: str = "float32"

    mse_epsilon_threshold: float = 0.01



from pydantic import ConfigDict, Field, field_validator
from pydantic.dataclasses import dataclass

from sae_bench.evals.absorption.eval_config import AbsorptionEvalConfig
from sae_bench.evals.base_eval_output import (
    DEFAULT_DISPLAY,
    BaseEvalOutput,
    BaseMetricCategories,
    BaseMetrics,
    BaseResultDetail,
)

EVAL_TYPE_ID_ABSORPTION = "absorption_first_letter"


# Define the metrics for each metric category, and include a title and description for each.
@dataclass
class AbsorptionMeanMetrics(BaseMetrics):
    mean_absorption_fraction_score: float = Field(
        title="Mean Absorption Fraction Score",
        description="Average of the absorption fraction scores across all letters",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    mean_full_absorption_score: float = Field(
        title="Mean Full Absorption Score",
        description="Average of the full absorption scores across all letters",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    mean_num_split_features: float = Field(
        title="Mean Number of Split Features",
        description="Average number of split features across all letters",
        json_schema_extra=DEFAULT_DISPLAY,
    )
    std_dev_absorption_fraction_score: float = Field(
        title="Standard Deviation of Absorption Fraction Score",
        description="Standard deviation of the absorption fraction scores across all letters",
    )
    std_dev_full_absorption_score: float = Field(
        title="Standard Deviation of Full Absorption Score",
        description="Standard deviation of the full absorption scores across all letters",
    )
    std_dev_num_split_features: float = Field(
        title="Standard Deviation of Number of Split Features",
        description="Standard deviation of the number of split features across all letters",
    )


# Define the categories themselves, and include a title and description for each.
@dataclass
class AbsorptionMetricCategories(BaseMetricCategories):
    mean: AbsorptionMeanMetrics = Field(
        title="Mean",
        description="Mean metrics",
        json_schema_extra=DEFAULT_DISPLAY,
    )


# Define a result detail, which in this case is an absorption result for a single letter.
@dataclass
class AbsorptionResultDetail(BaseResultDetail):
    first_letter: str = Field(title="First Letter", description="")

    @field_validator("first_letter")
    @classmethod
    def validate_single_letter(cls, value: str) -> str:
        if len(value) == 1 and value.isalpha():
            return value
        raise ValueError("First letter must be a single letter")

    mean_absorption_fraction: float = Field(
        title="Mean Absorption Fraction", description=""
    )
    full_absorption_rate: float = Field(title="Rate of Full Absorption", description="")
    num_full_absorption: int = Field(title="Num Full Absorption", description="")
    num_probe_true_positives: int = Field(
        title="Num Probe True Positives", description=""
    )
    num_split_features: int = Field(title="Num Split Features", description="")


# Define the eval output, which includes the eval config, metrics, and result details.
# The title will end up being the title of the eval in the UI.
@dataclass(config=ConfigDict(title="Absorption"))
class AbsorptionEvalOutput(
    BaseEvalOutput[
        AbsorptionEvalConfig, AbsorptionMetricCategories, AbsorptionResultDetail
    ]
):
    # This will end up being the description of the eval in the UI.
    """
    The feature absorption evaluation looking at the first letter.
    """

    eval_config: AbsorptionEvalConfig
    eval_id: str
    datetime_epoch_millis: int
    eval_result_metrics: AbsorptionMetricCategories
    eval_result_details: list[AbsorptionResultDetail] = Field(
        default_factory=list,
        title="Per-Letter Absorption Results",
        description="Each object is a stat on the first letter of the absorption.",
    )
    eval_type_id: str = Field(
        default=EVAL_TYPE_ID_ABSORPTION,
        title="Eval Type ID",
        description="The type of the evaluation",
    )



from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

from sae_bench.evals.base_eval_output import (
    DEFAULT_DISPLAY,
    BaseEvalOutput,
    BaseMetricCategories,
    BaseMetrics,
    BaseResultDetail,
)
from sae_bench.evals.unlearning.eval_config import UnlearningEvalConfig

EVAL_TYPE_ID_UNLEARNING = "unlearning"


@dataclass
class UnlearningMetrics(BaseMetrics):
    unlearning_score: float = Field(
        title="Unlearning Score",
        description="Unlearning score, using methodology from APPLYING SPARSE AUTOENCODERS TO UNLEARN KNOWLEDGE IN LANGUAGE MODELS",
        json_schema_extra=DEFAULT_DISPLAY,
    )


# Define the categories themselves
@dataclass
class UnlearningMetricCategories(BaseMetricCategories):
    unlearning: UnlearningMetrics = Field(
        title="Unlearning",
        description="Metrics related to unlearning",
    )


# Define the eval output
@dataclass(config=ConfigDict(title="Unlearning"))
class UnlearningEvalOutput(
    BaseEvalOutput[UnlearningEvalConfig, UnlearningMetricCategories, BaseResultDetail]
):
    """
    An evaluation of the ability of SAEs to unlearn biology knowledge from LLMs, using methodology from `Applying Sparse Autoencoders to Unlearn Knowledge in Language Models`
    """

    eval_config: UnlearningEvalConfig
    eval_id: str
    datetime_epoch_millis: int
    eval_result_metrics: UnlearningMetricCategories

    eval_type_id: str = Field(
        default=EVAL_TYPE_ID_UNLEARNING,
        title="Eval Type ID",
        description="The type of the evaluation",
    )



import argparse
import gc
import os
import random
import shutil
import time
from dataclasses import asdict
from datetime import datetime

import torch
from sae_lens import SAE
from tqdm import tqdm
from transformer_lens import HookedTransformer

import sae_bench.evals.sparse_probing.probe_training as probe_training
import sae_bench.sae_bench_utils.activation_collection as activation_collection
import sae_bench.sae_bench_utils.dataset_info as dataset_info
import sae_bench.sae_bench_utils.dataset_utils as dataset_utils
import sae_bench.sae_bench_utils.general_utils as general_utils
from sae_bench.evals.sparse_probing.eval_config import SparseProbingEvalConfig
from sae_bench.evals.sparse_probing.eval_output import (
    EVAL_TYPE_ID_SPARSE_PROBING,
    SparseProbingEvalOutput,
    SparseProbingLlmMetrics,
    SparseProbingMetricCategories,
    SparseProbingResultDetail,
    SparseProbingSaeMetrics,
)
from sae_bench.sae_bench_utils import (
    get_eval_uuid,
    get_sae_bench_version,
    get_sae_lens_version,
)
from sae_bench.sae_bench_utils.sae_selection_utils import (
    get_saes_from_regex,
)


def get_dataset_activations(
    dataset_name: str,
    config: SparseProbingEvalConfig,
    model: HookedTransformer,
    llm_batch_size: int,
    layer: int,
    hook_point: str,
    device: str,
) -> tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]:
    train_data, test_data = dataset_utils.get_multi_label_train_test_data(
        dataset_name,
        config.probe_train_set_size,
        config.probe_test_set_size,
        config.random_seed,
    )

    chosen_classes = dataset_info.chosen_classes_per_dataset[dataset_name]

    train_data = dataset_utils.filter_dataset(train_data, chosen_classes)
    test_data = dataset_utils.filter_dataset(test_data, chosen_classes)

    train_data = dataset_utils.tokenize_data_dictionary(
        train_data,
        model.tokenizer,  # type: ignore
        config.context_length,
        device,
    )
    test_data = dataset_utils.tokenize_data_dictionary(
        test_data,
        model.tokenizer,  # type: ignore
        config.context_length,
        device,
    )

    all_train_acts_BLD = activation_collection.get_all_llm_activations(
        train_data,
        model,
        llm_batch_size,
        layer,
        hook_point,
        mask_bos_pad_eos_tokens=True,
    )
    all_test_acts_BLD = activation_collection.get_all_llm_activations(
        test_data,
        model,
        llm_batch_size,
        layer,
        hook_point,
        mask_bos_pad_eos_tokens=True,
    )

    return all_train_acts_BLD, all_test_acts_BLD


def run_eval_single_dataset(
    dataset_name: str,
    config: SparseProbingEvalConfig,
    sae: SAE,
    model: HookedTransformer,
    layer: int,
    hook_point: str,
    device: str,
    artifacts_folder: str,
    save_activations: bool,
) -> tuple[dict[str, float], dict]:
    """config: eval_config.EvalConfig contains all hyperparameters to reproduce the evaluation.
    It is saved in the results_dict for reproducibility."""

    per_class_results_dict = {}

    activations_filename = f"{dataset_name}_activations.pt".replace("/", "_")

    activations_path = os.path.join(artifacts_folder, activations_filename)

    if not os.path.exists(activations_path):
        if config.lower_vram_usage:
            model = model.to(device)  # type: ignore
        all_train_acts_BLD, all_test_acts_BLD = get_dataset_activations(
            dataset_name,
            config,
            model,
            config.llm_batch_size,  # type: ignore
            layer,
            hook_point,
            device,
        )
        if config.lower_vram_usage:
            model = model.to("cpu")  # type: ignore

        all_train_acts_BD = activation_collection.create_meaned_model_activations(
            all_train_acts_BLD
        )

        all_test_acts_BD = activation_collection.create_meaned_model_activations(
            all_test_acts_BLD
        )

        # We use GPU here as sklearn.fit is slow on large input dimensions, all other probe training is done with sklearn.fit
        llm_probes, llm_test_accuracies = probe_training.train_probe_on_activations(
            all_train_acts_BD,
            all_test_acts_BD,
            select_top_k=None,
            use_sklearn=False,
            batch_size=250,
            epochs=100,
            lr=1e-2,
        )

        llm_results = {"llm_test_accuracy": llm_test_accuracies}

        for k in config.k_values:
            llm_top_k_probes, llm_top_k_test_accuracies = (
                probe_training.train_probe_on_activations(
                    all_train_acts_BD,
                    all_test_acts_BD,
                    select_top_k=k,
                )
            )
            llm_results[f"llm_top_{k}_test_accuracy"] = llm_top_k_test_accuracies

        acts = {
            "train": all_train_acts_BLD,
            "test": all_test_acts_BLD,
            "llm_results": llm_results,
        }

        if save_activations:
            torch.save(acts, activations_path)
    else:
        if config.lower_vram_usage:
            model = model.to("cpu")  # type: ignore
        print(f"Loading activations from {activations_path}")
        acts = torch.load(activations_path)
        all_train_acts_BLD = acts["train"]
        all_test_acts_BLD = acts["test"]
        llm_results = acts["llm_results"]

    all_sae_train_acts_BF = activation_collection.get_sae_meaned_activations(
        all_train_acts_BLD, sae, config.sae_batch_size
    )
    all_sae_test_acts_BF = activation_collection.get_sae_meaned_activations(
        all_test_acts_BLD, sae, config.sae_batch_size
    )

    for key in list(all_train_acts_BLD.keys()):
        del all_train_acts_BLD[key]
        del all_test_acts_BLD[key]

    if not config.lower_vram_usage:
        # This is optional, checking the accuracy of a probe trained on the entire SAE activations
        # We use GPU here as sklearn.fit is slow on large input dimensions, all other probe training is done with sklearn.fit
        _, sae_test_accuracies = probe_training.train_probe_on_activations(
            all_sae_train_acts_BF,
            all_sae_test_acts_BF,
            select_top_k=None,
            use_sklearn=False,
            batch_size=250,
            epochs=100,
            lr=1e-2,
        )
        per_class_results_dict["sae_test_accuracy"] = sae_test_accuracies
    else:
        per_class_results_dict["sae_test_accuracy"] = {"-1": -1}

        for key in all_sae_train_acts_BF.keys():
            all_sae_train_acts_BF[key] = all_sae_train_acts_BF[key].cpu()
            all_sae_test_acts_BF[key] = all_sae_test_acts_BF[key].cpu()

        torch.cuda.empty_cache()
        gc.collect()

    for llm_result_key, llm_result_value in llm_results.items():
        per_class_results_dict[llm_result_key] = llm_result_value

    for k in config.k_values:
        sae_top_k_probes, sae_top_k_test_accuracies = (
            probe_training.train_probe_on_activations(
                all_sae_train_acts_BF,
                all_sae_test_acts_BF,
                select_top_k=k,
            )
        )
        per_class_results_dict[f"sae_top_{k}_test_accuracy"] = sae_top_k_test_accuracies

    results_dict = {}
    for key, test_accuracies_dict in per_class_results_dict.items():
        average_test_acc = sum(test_accuracies_dict.values()) / len(
            test_accuracies_dict
        )
        results_dict[key] = average_test_acc

    return results_dict, per_class_results_dict


def run_eval_single_sae(
    config: SparseProbingEvalConfig,
    sae: SAE,
    model: HookedTransformer,
    device: str,
    artifacts_folder: str,
    save_activations: bool = True,
) -> tuple[dict[str, float | dict[str, float]], dict]:
    """hook_point: str is transformer lens format. example: f'blocks.{layer}.hook_resid_post'
    By default, we save activations for all datasets, and then reuse them for each sae.
    This is important to avoid recomputing activations for each SAE, and to ensure that the same activations are used for all SAEs.
    However, it can use 10s of GBs of disk space."""

    random.seed(config.random_seed)
    torch.manual_seed(config.random_seed)
    os.makedirs(artifacts_folder, exist_ok=True)

    results_dict = {}

    dataset_results = {}
    per_class_dict = {}
    for dataset_name in config.dataset_names:
        (
            dataset_results[f"{dataset_name}_results"],
            per_class_dict[f"{dataset_name}_results"],
        ) = run_eval_single_dataset(
            dataset_name,
            config,
            sae,
            model,
            sae.cfg.hook_layer,
            sae.cfg.hook_name,
            device,
            artifacts_folder,
            save_activations,
        )

    results_dict = general_utils.average_results_dictionaries(
        dataset_results, config.dataset_names
    )

    for dataset_name, dataset_result in dataset_results.items():
        results_dict[f"{dataset_name}"] = dataset_result

    if config.lower_vram_usage:
        model = model.to(device)  # type: ignore

    return results_dict, per_class_dict  # type: ignore


def run_eval(
    config: SparseProbingEvalConfig,
    selected_saes: list[tuple[str, SAE]] | list[tuple[str, str]],
    device: str,
    output_path: str,
    force_rerun: bool = False,
    clean_up_activations: bool = False,
    save_activations: bool = True,
    artifacts_path: str = "artifacts",
):
    """
    selected_saes is a list of either tuples of (sae_lens release, sae_lens id) or (sae_name, SAE object)

    If clean_up_activations is True, which means that the activations are deleted after the evaluation is done.
    You may want to use this because activations for all datasets can easily be 10s of GBs.
    Return dict is a dict of SAE name: evaluation results for that SAE."""
    eval_instance_id = get_eval_uuid()
    sae_lens_version = get_sae_lens_version()
    sae_bench_commit_hash = get_sae_bench_version()

    artifacts_folder = None
    os.makedirs(output_path, exist_ok=True)

    results_dict = {}

    llm_dtype = general_utils.str_to_dtype(config.llm_dtype)

    model = HookedTransformer.from_pretrained_no_processing(
        config.model_name, device=device, dtype=llm_dtype
    )

    for sae_release, sae_object_or_id in tqdm(
        selected_saes, desc="Running SAE evaluation on all selected SAEs"
    ):
        sae_id, sae, sparsity = general_utils.load_and_format_sae(
            sae_release, sae_object_or_id, device
        )  # type: ignore
        sae = sae.to(device=device, dtype=llm_dtype)

        sae_result_path = general_utils.get_results_filepath(
            output_path, sae_release, sae_id
        )

        if os.path.exists(sae_result_path) and not force_rerun:
            print(f"Skipping {sae_release}_{sae_id} as results already exist")
            continue

        artifacts_folder = os.path.join(
            artifacts_path,
            EVAL_TYPE_ID_SPARSE_PROBING,
            config.model_name,
            sae.cfg.hook_name,
        )

        sparse_probing_results, per_class_dict = run_eval_single_sae(
            config,
            sae,
            model,
            device,
            artifacts_folder,
            save_activations=save_activations,
        )
        eval_output = SparseProbingEvalOutput(
            eval_config=config,
            eval_id=eval_instance_id,
            datetime_epoch_millis=int(datetime.now().timestamp() * 1000),
            eval_result_metrics=SparseProbingMetricCategories(
                llm=SparseProbingLlmMetrics(
                    **{
                        k: v
                        for k, v in sparse_probing_results.items()
                        if k.startswith("llm_") and not isinstance(v, dict)
                    }
                ),
                sae=SparseProbingSaeMetrics(
                    **{
                        k: v
                        for k, v in sparse_probing_results.items()
                        if k.startswith("sae_") and not isinstance(v, dict)
                    }
                ),
            ),
            eval_result_details=[
                SparseProbingResultDetail(
                    dataset_name=dataset_name,
                    **result,
                )
                for dataset_name, result in sparse_probing_results.items()
                if isinstance(result, dict)
            ],
            eval_result_unstructured=per_class_dict,
            sae_bench_commit_hash=sae_bench_commit_hash,
            sae_lens_id=sae_id,
            sae_lens_release_id=sae_release,
            sae_lens_version=sae_lens_version,
            sae_cfg_dict=asdict(sae.cfg),
        )

        results_dict[f"{sae_release}_{sae_id}"] = asdict(eval_output)

        eval_output.to_json_file(sae_result_path, indent=2)

        gc.collect()
        torch.cuda.empty_cache()

    if clean_up_activations:
        if artifacts_folder is not None and os.path.exists(artifacts_folder):
            shutil.rmtree(artifacts_folder)

    return results_dict


def create_config_and_selected_saes(
    args,
) -> tuple[SparseProbingEvalConfig, list[tuple[str, str]]]:
    config = SparseProbingEvalConfig(
        model_name=args.model_name,
    )

    if args.llm_batch_size is not None:
        config.llm_batch_size = args.llm_batch_size
    else:
        config.llm_batch_size = activation_collection.LLM_NAME_TO_BATCH_SIZE[
            config.model_name
        ]

    if args.llm_dtype is not None:
        config.llm_dtype = args.llm_dtype
    else:
        config.llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

    if args.sae_batch_size is not None:
        config.sae_batch_size = args.sae_batch_size

    if args.random_seed is not None:
        config.random_seed = args.random_seed

    if args.lower_vram_usage:
        config.lower_vram_usage = True

    selected_saes = get_saes_from_regex(args.sae_regex_pattern, args.sae_block_pattern)
    assert len(selected_saes) > 0, "No SAEs selected"

    releases = set([release for release, _ in selected_saes])

    print(f"Selected SAEs from releases: {releases}")

    for release, sae in selected_saes:
        print(f"Sample SAEs: {release}, {sae}")

    return config, selected_saes


def arg_parser():
    parser = argparse.ArgumentParser(description="Run sparse probing evaluation")
    parser.add_argument("--random_seed", type=int, default=None, help="Random seed")
    parser.add_argument("--model_name", type=str, required=True, help="Model name")
    parser.add_argument(
        "--sae_regex_pattern",
        type=str,
        required=True,
        help="Regex pattern for SAE selection",
    )
    parser.add_argument(
        "--sae_block_pattern",
        type=str,
        required=True,
        help="Regex pattern for SAE block selection",
    )
    parser.add_argument(
        "--output_folder",
        type=str,
        default="eval_results/sparse_probing",
        help="Output folder",
    )
    parser.add_argument(
        "--force_rerun", action="store_true", help="Force rerun of experiments"
    )
    parser.add_argument(
        "--clean_up_activations",
        action="store_true",
        help="Clean up activations after evaluation",
    )
    parser.add_argument(
        "--save_activations",
        action="store_false",
        help="Save the generated LLM activations for later use",
    )
    parser.add_argument(
        "--llm_batch_size",
        type=int,
        default=None,
        help="Batch size for LLM. If None, will be populated using LLM_NAME_TO_BATCH_SIZE",
    )
    parser.add_argument(
        "--llm_dtype",
        type=str,
        default=None,
        choices=[None, "float32", "float64", "float16", "bfloat16"],
        help="Data type for LLM. If None, will be populated using LLM_NAME_TO_DTYPE",
    )
    parser.add_argument(
        "--sae_batch_size",
        type=int,
        default=None,
        help="Batch size for SAE. If None, will be populated using default config value",
    )
    parser.add_argument(
        "--lower_vram_usage",
        action="store_true",
        help="Lower GPU memory usage by doing more computation on the CPU. Useful on 1M width SAEs. Will be slower and require more system memory.",
    )
    parser.add_argument(
        "--artifacts_path",
        type=str,
        default="artifacts",
        help="Path to save artifacts",
    )

    return parser


if __name__ == "__main__":
    """
    python -m sae_bench.evals.sparse_probing.main \
    --sae_regex_pattern "sae_bench_pythia70m_sweep_standard_ctx128_0712" \
    --sae_block_pattern "blocks.4.hook_resid_post__trainer_10" \
    --model_name pythia-70m-deduped


    """
    args = arg_parser().parse_args()
    device = general_utils.setup_environment()

    start_time = time.time()

    config, selected_saes = create_config_and_selected_saes(args)

    print(selected_saes)

    # create output folder
    os.makedirs(args.output_folder, exist_ok=True)

    # run the evaluation on all selected SAEs
    results_dict = run_eval(
        config,
        selected_saes,
        device,
        args.output_folder,
        args.force_rerun,
        args.clean_up_activations,
        args.save_activations,
        artifacts_path=args.artifacts_path,
    )

    end_time = time.time()

    print(f"Finished evaluation in {end_time - start_time} seconds")


# Use this code snippet to use custom SAE objects
# if __name__ == "__main__":
#     import sae_bench.custom_saes.identity_sae as identity_sae
#     import sae_bench.custom_saes.jumprelu_sae as jumprelu_sae

#     """
#     python evals/sparse_probing/main.py
#     """
#     device = general_utils.setup_environment()

#     start_time = time.time()

#     random_seed = 42
#     output_folder = "eval_results/sparse_probing"

#     model_name = "gemma-2-2b"
#     hook_layer = 20

#     repo_id = "google/gemma-scope-2b-pt-res"
#     filename = f"layer_{hook_layer}/width_16k/average_l0_71/params.npz"
#     sae = jumprelu_sae.load_jumprelu_sae(repo_id, filename, hook_layer)
#     selected_saes = [(f"{repo_id}_{filename}_gemmascope_sae", sae)]

#     config = SparseProbingEvalConfig(
#         random_seed=random_seed,
#         model_name=model_name,
#     )

#     config.llm_batch_size = activation_collection.LLM_NAME_TO_BATCH_SIZE[config.model_name]
#     config.llm_dtype = activation_collection.LLM_NAME_TO_DTYPE[config.model_name]

#     # create output folder
#     os.makedirs(output_folder, exist_ok=True)

#     # run the evaluation on all selected SAEs
#     results_dict = run_eval(
#         config,
#         selected_saes,
#         device,
#         output_folder,
#         force_rerun=True,
#         clean_up_activations=False,
#         save_activations=True,
#     )

#     end_time = time.time()

#     print(f"Finished evaluation in {end_time - start_time} seconds")



import torch
from jaxtyping import Float
from sae_lens import SAE
from torch import Tensor
from transformer_lens.hook_points import HookPoint


def anthropic_clamp_resid_SAE_features(
    resid: Float[Tensor, "batch seq d_model"],
    hook: HookPoint,
    sae: SAE,
    features_to_ablate: list[int],
    multiplier: float = 1.0,
    random: bool = False,
) -> Float[Tensor, "batch seq d_model"] | None:
    """
    Given a list of feature indices, this hook function removes feature activations in a manner similar to the one
    used in "Scaling Monosemanticity": https://transformer-circuits.pub/2024/scaling-monosemanticity/index.html#appendix-methods-steering
    This version clamps the feature activation to the value(s) specified in multiplier
    """

    if len(features_to_ablate) > 0:
        with torch.no_grad():
            # adjust feature activations with scaling (multiplier = 0 just ablates the feature)
            feature_activations = sae.encode(resid)

            feature_activations[:, 0, :] = (
                0.0  # We zero out the BOS token for all SAEs.
            )
            # We don't need to zero out padding tokens because we right pad, so they don't effect the model generation.

            reconstruction = sae.decode(feature_activations)
            # else:
            #     try:
            #         import sys
            #         sys.path.append('/root')

            #         from dictionary_learning import AutoEncoder
            #         from dictionary_learning.trainers.top_k import AutoEncoderTopK

            #         if isinstance(sae, (AutoEncoder, AutoEncoderTopK)):
            #             reconstruction = sae(resid)
            #             feature_activations = sae.encode(resid)
            #     except:
            #         raise ValueError("sae must be an instance of SparseAutoencoder or SAE")

            error = resid - reconstruction

            non_zero_features_BLD = feature_activations[:, :, features_to_ablate] > 0

            # B, L, _ = non_zero_features_BLD.shape

            # non_zero_features_BD = non_zero_features_BLD.any(dim=1)
            # non_zero_features_BLD = einops.repeat(non_zero_features_BD, "B D -> B L D", L=L)

            if not random:
                if isinstance(multiplier, float) or isinstance(multiplier, int):
                    feature_activations[:, :, features_to_ablate] = torch.where(
                        non_zero_features_BLD,
                        -multiplier,
                        feature_activations[:, :, features_to_ablate],
                    )
                else:
                    raise NotImplementedError("Currently deprecated")
                    feature_activations[:, :, features_to_ablate] = torch.where(
                        non_zero_features_BLD,
                        -multiplier.unsqueeze(dim=0).unsqueeze(dim=0),
                        feature_activations[:, :, features_to_ablate],
                    )

            # set the next feature id's activations to the multiplier only if the previous feature id's
            # activations are positive
            else:
                raise NotImplementedError("Currently deprecated")
                assert isinstance(multiplier, float) or isinstance(multiplier, int)

                next_features_to_ablate = [
                    (f + 1) % feature_activations.shape[-1] for f in features_to_ablate
                ]
                feature_activations[:, :, next_features_to_ablate] = torch.where(
                    feature_activations[:, :, features_to_ablate] > 0,
                    -multiplier,
                    feature_activations[:, :, next_features_to_ablate],
                )

            # try:
            #     modified_reconstruction = (
            #         einops.einsum(
            #             feature_activations, sae.W_dec, "... d_sae, d_sae d_in -> ... d_in"
            #         )
            #         + sae.b_dec
            #     )
            # except:
            # SAEBench doesn't have W_dec and b_dec
            modified_reconstruction = sae.decode(feature_activations)

            # Unscale outputs if needed:
            # if sae.input_scaling_factor is not None:
            #     modified_reconstruction = modified_reconstruction / sae.input_scaling_factor
            resid = error + modified_reconstruction
        return resid
    return None
